# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Generate T2* map from multi-echo BOLD images
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_bold_t2s_wf

"""

import typing as ty

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe

from nibabies import config
from nibabies.interfaces.maths import Clip, Label2Mask
from nibabies.interfaces.multiecho import T2SMap
from nibabies.interfaces.reports import LabeledHistogram

LOGGER = config.loggers.workflow


# pylint: disable=R0914
def init_bold_t2s_wf(
    echo_times: ty.Sequence[float],
    mem_gb: float,
    omp_nthreads: int,
    me_t2s_fit_method: ty.Literal['curvefit', 'loglin'] = 'curvefit',
    name: str = 'bold_t2s_wf',
):
    r"""
    Combine multiple echos of :abbr:`ME-EPI (multi-echo echo-planar imaging)`.

    This workflow wraps the `tedana`_ :func:`T2* workflow <tedana.workflows.t2smap_workflow>`
    to optimally combine multiple preprocessed echos and derive a T2\ :sup:`★` map.
    The following steps are performed:
    #. Compute the T2\ :sup:`★` map
    #. Create an optimally combined ME-EPI time series

    .. _tedana: https://github.com/me-ica/tedana

    Parameters
    ----------
    echo_times : :obj:`list` or :obj:`tuple`
        list of TEs associated with each echo
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    omp_nthreads : :obj:`int`
        Maximum number of threads an individual process may use
    name : :obj:`str`
        Name of workflow (default: ``bold_t2s_wf``)

    Inputs
    ------
    bold_file
        list of individual echo files
    bold_mask
        a binary mask to apply to the BOLD files

    Outputs
    -------
    bold
        the optimally combined time series for all supplied echos
    t2star_map
        the calculated T2\ :sup:`★` map

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.morphology import BinaryDilation

    workflow = Workflow(name=name)
    if me_t2s_fit_method == 'curvefit':
        fit_str = (
            'nonlinear regression. '
            'The T2<sup>★</sup>/S<sub>0</sub> estimates from a log-linear regression fit '
            'were used for initial values'
        )
    elif me_t2s_fit_method == 'loglin':
        fit_str = 'log-linear regression'
    else:
        fit_str = f'unknown method: {me_t2s_fit_method}'

    workflow.__desc__ = f"""\
A T2<sup>★</sup> map was estimated from the preprocessed EPI echoes, by voxel-wise fitting
the maximal number of echoes with reliable signal in that voxel to a monoexponential signal
decay model with {fit_str}.
The calculated T2<sup>★</sup> map was then used to optimally combine preprocessed BOLD across
echoes following the method described in [@posse_t2s].
The optimally combined time series was carried forward as the *preprocessed BOLD*.
"""

    inputnode = pe.Node(niu.IdentityInterface(fields=['bold_file', 'bold_mask']), name='inputnode')

    outputnode = pe.Node(niu.IdentityInterface(fields=['bold', 't2star_map']), name='outputnode')

    LOGGER.log(25, 'Generating T2* map and optimally combined ME-EPI time series.')

    dilate_mask = pe.Node(BinaryDilation(radius=2), name='dilate_mask')

    t2smap_node = pe.Node(
        T2SMap(echo_times=list(echo_times), fittype=me_t2s_fit_method),
        name='t2smap_node',
        mem_gb=2.5 * mem_gb * len(echo_times),
    )
    workflow.connect([
        (inputnode, dilate_mask, [('bold_mask', 'in_mask')]),
        (inputnode, t2smap_node, [('bold_file', 'in_files')]),
        (dilate_mask, t2smap_node, [('out_mask', 'mask_file')]),
        (t2smap_node, outputnode, [('optimal_comb', 'bold'),
                                   ('t2star_map', 't2star_map')]),
    ])  # fmt:skip

    return workflow


def init_t2s_reporting_wf(name: str = 't2s_reporting_wf'):
    r"""
    Generate T2\*-map reports.

    This workflow generates a histogram of estimated T2\* values (in seconds) in the
    cortical and subcortical gray matter mask.

    Parameters
    ----------
    mem_gb : :obj:`float`
        Size of BOLD file in GB
    omp_nthreads : :obj:`int`
        Maximum number of threads an individual process may use
    name : :obj:`str`
        Name of workflow (default: ``t2s_reporting_wf``)

    Inputs
    ------
    t2star_file
        estimated T2\* map
    boldref
        reference BOLD file
    label_file
        an integer label file identifying gray matter with value ``1``
    boldref2anat_xfm
        Affine matrix that maps images in the native bold space into the
        anatomical space of ``label_file``; can be ``"identity"`` if label
        file is already aligned

    Outputs
    -------
    t2star_hist
        an SVG histogram showing estimated T2\* values in gray matter
    t2s_comp_report
        a before/after figure comparing the reference BOLD image and T2\* map
    """
    from nipype.pipeline import engine as pe
    from nireports.interfaces.reporting.base import (
        SimpleBeforeAfterRPT as SimpleBeforeAfter,
    )
    from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms

    workflow = pe.Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['t2star_file', 'boldref', 'label_file', 'boldref2anat_xfm']),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['t2star_hist', 't2s_comp_report']), name='outputnode'
    )

    label_tfm = pe.Node(
        ApplyTransforms(interpolation='MultiLabel', invert_transform_flags=[True]),
        name='label_tfm',
    )

    gm_mask = pe.Node(Label2Mask(label_val=1), name='gm_mask')

    clip_t2star = pe.Node(Clip(maximum=0.1), name='clip_t2star')

    t2s_hist = pe.Node(
        LabeledHistogram(mapping={1: 'Gray matter'}, xlabel='T2* (s)'), name='t2s_hist'
    )

    t2s_comparison = pe.Node(
        SimpleBeforeAfter(
            before_label='BOLD Reference',
            after_label='T2* Map',
            dismiss_affine=True,
        ),
        name='t2s_comparison',
        mem_gb=0.1,
    )
    workflow.connect([
        (inputnode, label_tfm, [('label_file', 'input_image'),
                                ('t2star_file', 'reference_image'),
                                ('boldref2anat_xfm', 'transforms')]),
        (inputnode, clip_t2star, [('t2star_file', 'in_file')]),
        (clip_t2star, t2s_hist, [('out_file', 'in_file')]),
        (label_tfm, gm_mask, [('output_image', 'in_file')]),
        (gm_mask, t2s_hist, [('out_file', 'label_file')]),
        (inputnode, t2s_comparison, [('boldref', 'before'),
                                     ('t2star_file', 'after')]),
        (gm_mask, t2s_comparison, [('out_file', 'wm_seg')]),
        (t2s_hist, outputnode, [('out_report', 't2star_hist')]),
        (t2s_comparison, outputnode, [('out_report', 't2s_comp_report')]),
    ])  # fmt:skip
    return workflow



"""py.test configuration"""

import json
from pathlib import Path
from shutil import copytree

import nibabel as nb
import numpy as np
import pytest

from nibabies.data import load as load_data

try:
    from importlib.resources import files as ir_files
except ImportError:  # PY<3.9
    from importlib_resources import files as ir_files


def copytree_or_skip(source, target):
    data_dir = ir_files('nibabies') / source
    if not data_dir.exists():
        pytest.skip(f'Cannot chdir into {data_dir!r}. Probably in a zipped distribution.')

    try:
        copytree(data_dir, target / data_dir.name)
    except Exception:  # noqa: BLE001
        pytest.skip(f'Cannot copy {data_dir!r} into {target / data_dir.name}. Probably in a zip.')


@pytest.fixture(autouse=True)
def _populate_namespace(doctest_namespace, tmp_path):
    doctest_namespace['copytree_or_skip'] = copytree_or_skip
    doctest_namespace['testdir'] = tmp_path
    doctest_namespace['datadir'] = load_data()


@pytest.fixture
def minimal_bids(tmp_path):
    bids = tmp_path / 'bids'
    bids.mkdir()
    Path.write_text(
        bids / 'dataset_description.json', json.dumps({'Name': 'Test DS', 'BIDSVersion': '1.8.0'})
    )
    T1w = bids / 'sub-01' / 'anat' / 'sub-01_T1w.nii.gz'
    T1w.parent.mkdir(parents=True)
    nb.Nifti1Image(np.zeros((5, 5, 5)), np.eye(4)).to_filename(T1w)
    return bids



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""This module provides interfaces for workbench surface commands"""

import os

from nipype import logging
from nipype.interfaces.base import CommandLineInputSpec, File, TraitedSpec, traits
from nipype.interfaces.workbench.base import WBCommand

iflogger = logging.getLogger('nipype.interface')


class MetricDilateInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr='%s ',
        position=0,
        desc='the metric to dilate',
    )

    surf_file = File(
        exists=True,
        mandatory=True,
        argstr='%s ',
        position=1,
        desc='the surface to compute on',
    )

    distance = traits.Float(
        mandatory=True,
        argstr='%f ',
        position=2,
        desc='distance in mm to dilate',
    )

    out_file = File(
        name_source=['in_file'],
        name_template='%s.func.gii',
        keep_extension=False,
        argstr='%s ',
        position=3,
        desc='output - the output metric',
    )

    bad_vertex_roi_file = File(
        argstr='-bad-vertex-roi %s ',
        position=4,
        desc='metric file, positive values denote vertices to have their values replaced',
    )

    data_roi_file = File(
        argstr='-data-roi %s ',
        position=5,
        desc='metric file, positive values denote vertices that have data',
    )

    column = traits.Int(
        position=6,
        argstr='-column %d ',
        desc='the column number',
    )

    nearest = traits.Bool(
        position=7,
        argstr='-nearest ',
        desc='use the nearest good value instead of a weighted average',
    )

    linear = traits.Bool(
        position=8,
        argstr='-linear ',
        desc='fill in values with linear interpolation along strongest gradient',
    )

    exponent = traits.Float(
        argstr='-exponent %f ',
        position=9,
        default=6.0,
        desc='exponent n to use in (area / (distance ^ n)) as the weighting function (default 6)',
    )

    corrected_areas = File(
        argstr='-corrected-areas %s ',
        position=10,
        desc='vertex areas to use instead of computing them from the surface',
    )

    legacy_cutoff = traits.Bool(
        position=11,
        argstr='-legacy-cutoff ',
        desc='use the v1.3.2 method of choosing how many vertices to '
        'use when calculating the dilated value with weighted method',
    )


class MetricDilateOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='output file')


class MetricDilate(WBCommand):
    """
    For all data values designated as bad, if they neighbor a good value or
    are within the specified distance of a good value in the same kind of
    model, replace the value with a distance weighted average of nearby good
    values, otherwise set the value to zero.  If -nearest is specified, it
    will use the value from the closest good value within range instead of a
    weighted average.  When the input file contains label data, nearest
    dilation is used on the surface, and weighted popularity is used in the
    volume.

    The -corrected-areas options are intended for dilating on group average
    surfaces, but it is only an approximate correction for the reduction of
    structure in a group average surface.

    If -bad-vertex-roi is specified, all values, including those with
    value zero, are good, except for locations with a positive value in the
    ROI.  If it is not specified, only values equal to zero are bad.
    """

    input_spec = MetricDilateInputSpec
    output_spec = MetricDilateOutputSpec
    _cmd = 'wb_command -metric-dilate '


class MetricResampleInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr='%s',
        position=0,
        desc='The metric file to resample',
    )
    current_sphere = File(
        exists=True,
        mandatory=True,
        argstr='%s',
        position=1,
        desc='A sphere surface with the mesh that the metric is currently on',
    )
    new_sphere = File(
        exists=True,
        mandatory=True,
        argstr='%s',
        position=2,
        desc='A sphere surface that is in register with <current-sphere> and'
        ' has the desired output mesh',
    )
    method = traits.Enum(
        'ADAP_BARY_AREA',
        'BARYCENTRIC',
        argstr='%s',
        mandatory=True,
        position=3,
        desc='The method name - ADAP_BARY_AREA method is recommended for'
        ' ordinary metric data, because it should use all data while'
        ' downsampling, unlike BARYCENTRIC. If ADAP_BARY_AREA is used,'
        ' exactly one of area_surfs or area_metrics must be specified',
    )
    out_file = File(
        name_source=['new_sphere'],
        name_template='%s.out',
        keep_extension=True,
        argstr='%s',
        position=4,
        desc='The output metric',
    )
    area_surfs = traits.Bool(
        position=5,
        argstr='-area-surfs',
        xor=['area_metrics'],
        desc='Specify surfaces to do vertex area correction based on',
    )
    area_metrics = traits.Bool(
        position=5,
        argstr='-area-metrics',
        xor=['area_surfs'],
        desc='Specify vertex area metrics to do area correction based on',
    )
    current_area = File(
        exists=True,
        position=6,
        argstr='%s',
        desc='A relevant anatomical surface with <current-sphere> mesh OR'
        ' a metric file with vertex areas for <current-sphere> mesh',
    )
    new_area = File(
        exists=True,
        position=7,
        argstr='%s',
        desc='A relevant anatomical surface with <current-sphere> mesh OR'
        ' a metric file with vertex areas for <current-sphere> mesh',
    )
    roi_metric = File(
        exists=True,
        position=8,
        argstr='-current-roi %s',
        desc='Input roi on the current mesh used to exclude non-data vertices',
    )
    valid_roi_out = traits.Bool(
        position=9,
        argstr='-valid-roi-out',
        desc='Output the ROI of vertices that got data from valid source vertices',
    )
    largest = traits.Bool(
        position=10,
        argstr='-largest',
        desc='Use only the value of the vertex with the largest weight',
    )


class MetricResampleOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output metric')
    roi_file = File(desc='ROI of vertices that got data from valid source vertices')


class MetricResample(WBCommand):
    """
    Resample a metric file to a different mesh

    Resamples a metric file, given two spherical surfaces that are in
    register.  If ``ADAP_BARY_AREA`` is used, exactly one of -area-surfs or
    ``-area-metrics`` must be specified.

    The ``ADAP_BARY_AREA`` method is recommended for ordinary metric data,
    because it should use all data while downsampling, unlike ``BARYCENTRIC``.
    The recommended areas option for most data is individual midthicknesses
    for individual data, and averaged vertex area metrics from individual
    midthicknesses for group average data.

    The ``-current-roi`` option only masks the input, the output may be slightly
    dilated in comparison, consider using ``-metric-mask`` on the output when
    using ``-current-roi``.

    The ``-largest option`` results in nearest vertex behavior when used with
    ``BARYCENTRIC``.  When resampling a binary metric, consider thresholding at
    0.5 after resampling rather than using ``-largest``.
    """

    input_spec = MetricResampleInputSpec
    output_spec = MetricResampleOutputSpec
    _cmd = 'wb_command -metric-resample'

    def _format_arg(self, opt, spec, val):
        if opt in ['current_area', 'new_area']:
            if not self.inputs.area_surfs and not self.inputs.area_metrics:
                raise ValueError(f'{opt} was set but neither area_surfs or area_metrics were set')
        if opt == 'method':
            if (
                val == 'ADAP_BARY_AREA'
                and not self.inputs.area_surfs
                and not self.inputs.area_metrics
            ):
                raise ValueError('Exactly one of area_surfs or area_metrics must be specified')
        if opt == 'valid_roi_out' and val:
            # generate a filename and add it to argstr
            roi_out = self._gen_filename(self.inputs.in_file, suffix='_roi')
            iflogger.info('Setting roi output file as', roi_out)
            spec.argstr += ' ' + roi_out
        return super()._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = super()._list_outputs()
        if self.inputs.valid_roi_out:
            roi_file = self._gen_filename(self.inputs.in_file, suffix='_roi')
            outputs['roi_file'] = os.path.abspath(roi_file)
        return outputs



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Miscellaneous utilities."""

from __future__ import annotations

from functools import cache
from pathlib import Path


def fix_multi_source_name(in_files):
    """
    Make up a generic source name when there are multiple
    >>> fix_multi_source_name([
    ...     '/path/to/sub-045_ses-test_T1w.nii.gz',
    ...     '/path/to/sub-045_ses-retest_T1w.nii.gz'])
    '/path/to/sub-045_T1w.nii.gz'
    """
    import re
    from pathlib import Path

    from nipype.utils.filemanip import filename_to_list

    if not isinstance(in_files, tuple | list):
        return in_files
    elif len(in_files) == 1:
        return in_files[0]

    p = Path(filename_to_list(in_files)[0])
    # subject_label = p.name.split("_", 1)[0].split("-")[1]
    try:
        subj = re.search(r'(?<=^sub-)[a-zA-Z0-9]*', p.name).group()
        suffix = re.search(r'(?<=_)\w+(?=\.)', p.name).group()
    except AttributeError as e:
        raise AttributeError('Could not extract BIDS information') from e
    return str(p.parent / f'sub-{subj}_{suffix}.nii.gz')


def check_deps(workflow):
    """Make sure dependencies are present in this system."""
    from nipype.utils.filemanip import which

    return sorted(
        (node.interface.__class__.__name__, node.interface._cmd)
        for node in workflow._get_all_nodes()
        if (hasattr(node.interface, '_cmd') and which(node.interface._cmd.split()[0]) is None)
    )


def cohort_by_months(template, months):
    """
    Produce a recommended cohort based on partipants age
    """
    cohort_key = {
        'MNIInfant': (
            # upper bound of template | cohort
            2,  # 1
            5,  # 2
            8,  # 3
            11,  # 4
            14,  # 5
            17,  # 6
            21,  # 7
            27,  # 8
            33,  # 9
            44,  # 10
            60,  # 11
        ),
        'UNCInfant': (
            8,  # 1
            12,  # 2
            24,  # 3
        ),
    }
    ages = cohort_key.get(template)
    if ages is None:
        raise KeyError('Template cohort information does not exist.')

    for cohort, age in enumerate(ages, 1):
        if months <= age:
            return cohort
    raise KeyError('Age exceeds all cohorts!')


def check_total_memory(recommended_gb):
    """
    Check total memory allocated to the process, and compare with a recommended value.
    If available memory is equal to or greater than recommended, return ``True``.
    Otherwise, return ``False``.
    """

    try:
        import psutil
    except ImportError:
        return

    tot = int(psutil.virtual_memory().total / 1024**3)
    return tot >= recommended_gb


def combine_meepi_source(in_files):
    """
    Create a new source name when optimally
    combining multiple multi-echo EPIs
    >>> combine_meepi_source([
    ...     'sub-01_run-01_echo-1_bold.nii.gz',
    ...     'sub-01_run-01_echo-2_bold.nii.gz',
    ...     'sub-01_run-01_echo-3_bold.nii.gz',])
    'sub-01_run-01_bold.nii.gz'
    """
    import os

    from nipype.utils.filemanip import filename_to_list

    base, in_file = os.path.split(filename_to_list(in_files)[0])
    entities = [ent for ent in in_file.split('_') if not ent.startswith('echo-')]
    basename = '_'.join(entities)
    return os.path.join(base, basename)


def get_file(pkg: str, src_path: str | Path) -> str:
    """
    Get or extract a source file.
    Assures the file will be available until the lifetime of the current Python process.
    """
    import atexit
    from contextlib import ExitStack

    try:
        from importlib.resources import as_file, files
    except ImportError:
        from importlib_resources import as_file, files

    file_manager = ExitStack()
    atexit.register(file_manager.close)
    ref = files(pkg) / str(src_path)
    fl = file_manager.enter_context(as_file(ref))
    return str(fl)


@cache
def estimate_bold_mem_usage(bold_fname: str) -> tuple[int, dict]:
    import nibabel as nb
    import numpy as np

    img = nb.load(bold_fname)
    nvox = int(np.prod(img.shape, dtype='u8'))
    # Assume tools will coerce to 8-byte floats to be safe
    bold_size_gb = 8 * nvox / (1024**3)
    bold_tlen = img.shape[-1]
    mem_gb = {
        'filesize': bold_size_gb,
        'resampled': bold_size_gb * 4,
        'largemem': bold_size_gb * (max(bold_tlen / 100, 1.0) + 4),
    }
    return bold_tlen, mem_gb



import os
import re

from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    InputMultiObject,
    OutputMultiObject,
    SimpleInterface,
    TraitedSpec,
    traits,
)


class CiftiSelectInputSpec(BaseInterfaceInputSpec):
    hemi = traits.Enum('L', 'R', desc='Hemisphere')
    surfaces = InputMultiObject(File(exists=True), desc='Surfaces')
    morphometrics = InputMultiObject(File(exists=True), desc='Surface morphometrics')
    spherical_registrations = InputMultiObject(
        File(exists=True), desc='Spherical registration to fsLR'
    )
    template_spheres = InputMultiObject(File(exists=True), desc='fsLR sphere')
    template_surfaces = InputMultiObject(File(exists=True), desc='fsLR midthickness')
    template_rois = InputMultiObject(File(exists=True), desc='fsLR ROIs')


class CiftiSelectOutputSpec(TraitedSpec):
    white = OutputMultiObject(File, desc='white surface')
    pial = OutputMultiObject(File, desc='pial surface')
    midthickness = OutputMultiObject(File, desc='midthickness surface')
    thickness = OutputMultiObject(File, desc='thickness surface')
    sphere_reg = OutputMultiObject(File, desc='fsLR spherical registration')
    template_sphere = OutputMultiObject(File, desc='fsLR sphere')
    template_surface = OutputMultiObject(File, desc='fsLR surface (midthickness)')
    template_roi = OutputMultiObject(File, desc='fsLR ROIs')


class CiftiSelect(SimpleInterface):
    input_spec = CiftiSelectInputSpec
    output_spec = CiftiSelectOutputSpec

    def _run_interface(self, runtime):
        idx = 0 if self.inputs.hemi == 'L' else 1
        all_surfaces = (self.inputs.surfaces or []) + (self.inputs.morphometrics or [])
        container = {
            'white': [],
            'pial': [],
            'midthickness': [],
            'thickness': [],
            'sphere_reg': self.inputs.spherical_registrations or [],
            'template_sphere': self.inputs.template_spheres or [],
            'template_surface': self.inputs.template_surfaces or [],
            'template_roi': self.inputs.template_rois or [],
        }
        find_name = re.compile(r'(?:^|[^d])(?P<name>white|pial|midthickness|thickness)')
        for surface in all_surfaces:
            match = find_name.search(os.path.basename(surface))
            if match:
                container[match.group('name')].append(surface)

        for name, vals in container.items():
            if vals:
                self._results[name] = sorted(vals, key=os.path.basename)[idx]
        return runtime



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# STATEMENT OF CHANGES: This file is derived from sources licensed under the Apache-2.0 terms,
# and this file has been changed.
# The original file this work derives from is found at:
# https://github.com/nipreps/fmriprep/blob/a4fd718/fmriprep/workflows/bold/base.py
#
# [January 2022] CHANGES:
#   * `init_nibabies_wf` now takes in a dictionary composed of participant/session key/values.
#   * `init_single_subject_wf` now differentiates between participant sessions.
#     This change is to treat sessions as a "first-class" identifier, to better handle the
#     potential rapid changing of brain morphometry.
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
NiBabies base processing workflows
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_nibabies_wf
.. autofunction:: init_single_subject_wf

"""

from __future__ import annotations

import os
import pprint
import sys
import typing as ty
import warnings
from copy import deepcopy

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.interfaces.utility import KeySelect
from niworkflows.utils.connections import listify
from packaging.version import Version
from smriprep.workflows.outputs import init_template_iterator_wf

from nibabies import config
from nibabies.interfaces import DerivativesDataSink
from nibabies.interfaces.reports import AboutSummary, SubjectSummary
from nibabies.utils.bids import parse_bids_for_age_months
from nibabies.workflows.anatomical.apply import init_infant_anat_apply_wf
from nibabies.workflows.anatomical.fit import (
    init_infant_anat_fit_wf,
    init_infant_single_anat_fit_wf,
)
from nibabies.workflows.bold.base import init_bold_wf

if ty.TYPE_CHECKING:
    from bids.layout import BIDSLayout
    from niworkflows.utils.spaces import SpatialReferences

    SubjectSession = tuple[str, str | None]


MCRIBS_RECOMMEND_AGE_CAP = 3


def init_nibabies_wf(subworkflows_list: list[SubjectSession]):
    """
    Build *NiBabies*'s pipeline.

    This workflow organizes the execution of NiBabies, with a sub-workflow for
    each subject.

    If FreeSurfer's ``infant_recon_all`` is to be run, a corresponding folder is created
    and populated with any needed template subjects under the derivatives folder.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.tests import mock_config
            from nibabies.workflows.base import init_nibabies_wf
            with mock_config():
                wf = init_nibabies_wf(['01', None])

    Parameters
    ----------
    subworkflows_list: :obj:`list` of :obj:`tuple`
        A list of the subworkflows to create.
        Each subject session is run as an individual workflow.
    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.bids import BIDSFreeSurferDir

    ver = Version(config.environment.version)
    nibabies_wf = Workflow(name=f'nibabies_{ver.major}_{ver.minor}_wf')
    nibabies_wf.base_dir = config.execution.work_dir

    execution_spaces = init_execution_spaces()
    surface_recon = config.workflow.surface_recon_method is not None
    if surface_recon:
        fsdir = pe.Node(
            BIDSFreeSurferDir(
                derivatives=config.execution.output_dir,
                freesurfer_home=os.getenv('FREESURFER_HOME'),
                spaces=execution_spaces.get_fs_spaces(),
            ),
            name=f'fsdir_run_{config.execution.run_uuid.replace("-", "_")}',
            run_without_submitting=True,
        )
        if config.execution.fs_subjects_dir is not None:
            fsdir.inputs.subjects_dir = str(config.execution.fs_subjects_dir.absolute())

    for subject_id, session_id in subworkflows_list:
        # Calculate the age and age-specific spaces
        age = parse_bids_for_age_months(config.execution.bids_dir, subject_id, session_id)
        if config.workflow.age_months:
            config.loggers.cli.warning(
                '`--age-months` is deprecated and will be removed in a future release.'
                'Please use a `sessions.tsv` or `participants.tsv` file to track participants age.'
            )
            age = config.workflow.age_months
        if age is None:
            raise RuntimeError(
                'Could not find age for sub-{subject}{session}'.format(
                    subject=subject_id, session=f'_ses-{session_id}' if session_id else ''
                )
            )
        output_spaces = init_workflow_spaces(execution_spaces, age)

        bids_level = [f'sub-{subject_id}']
        if session_id is not None:
            bids_level.append(f'ses-{session_id}')

        log_dir = (
            config.execution.nibabies_dir.joinpath(*bids_level) / 'log' / config.execution.run_uuid
        )
        # skull strip template cohort
        single_subject_wf = init_single_subject_wf(
            subject_id=subject_id,
            age=age,
            session_id=session_id,
            spaces=output_spaces,
        )

        single_subject_wf.config['execution']['crashdump_dir'] = str(log_dir)
        for node in single_subject_wf._get_all_nodes():
            node.config = deepcopy(single_subject_wf.config)
        if surface_recon:
            nibabies_wf.connect(fsdir, 'subjects_dir', single_subject_wf, 'inputnode.subjects_dir')
        else:
            nibabies_wf.add_nodes([single_subject_wf])

        # Dump a copy of the config file into the log directory
        log_dir.mkdir(exist_ok=True, parents=True)
        config.to_filename(log_dir / 'nibabies.toml')

    return nibabies_wf


def init_single_subject_wf(
    *,
    subject_id: str,
    age: int,
    session_id: str | None = None,
    spaces: SpatialReferences | None = None,
):
    """
    Organize the preprocessing pipeline for a single subject, at a single session.

    It collects and reports information about the subject, and prepares
    sub-workflows to perform anatomical and functional preprocessing.
    Anatomical preprocessing is performed in a single workflow, regardless of
    the number of sessions.
    Functional preprocessing is performed using a separate workflow for each
    individual BOLD series.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from nibabies.workflows.tests import mock_config
            from nibabies.workflows.base import init_single_subject_wf
            with mock_config():
                wf = init_single_subject_wf('01')

    Parameters
    ----------
    subject_id : :obj:`str`
        Subject label for this single-subject workflow.
    session_id : :obj:`str` or None
        Session identifier.
    age: :obj:`int` or None
        Age (in months) of subject.

    Inputs
    ------
    subjects_dir : :obj:`str`
        FreeSurfer's ``$SUBJECTS_DIR``.

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.bids import BIDSDataGrabber, BIDSInfo
    from niworkflows.interfaces.nilearn import NILEARN_VERSION
    from niworkflows.utils.bids import collect_data
    from niworkflows.utils.spaces import Reference

    from ..utils.misc import fix_multi_source_name

    subject_session_id = _subject_session_id(subject_id, session_id)
    workflow = Workflow(name=f'single_subject_{subject_session_id}_wf')
    workflow.__desc__ = f"""
Results included in this manuscript come from preprocessing
performed using *NiBabies* {config.environment.version},
derived from fMRIPrep (@fmriprep1; @fmriprep2; RRID:SCR_016216).
The underlying workflow engine used is *Nipype* {config.environment.nipype_version}
(@nipype1; @nipype2; RRID:SCR_002502).

"""
    workflow.__postdesc__ = f"""

Many internal operations of *NiBabies* use
*Nilearn* {NILEARN_VERSION} [@nilearn, RRID:SCR_001362],
mostly within the functional processing workflow.
For more details of the pipeline, see [the section corresponding
to workflows in *nibabies*'s documentation]\
(https://nibabies.readthedocs.io/en/latest/workflows.html \
"NiBabies's documentation").


### Copyright Waiver

The above boilerplate text was automatically generated by NiBabies
with the express intention that users should copy and paste this
text into their manuscripts *unchanged*.
It is released under the [CC0]\
(https://creativecommons.org/publicdomain/zero/1.0/) license.

### References

"""

    subject_data = collect_data(
        config.execution.layout,
        subject_id,
        session_id=session_id,
        task=config.execution.task_id,
        echo=config.execution.echo_idx,
        bids_filters=config.execution.bids_filters,
    )[0]

    if 'flair' in config.workflow.ignore:
        subject_data['flair'] = []
    if 't1w' in config.workflow.ignore:
        subject_data['t1w'] = []
    if 't2w' in config.workflow.ignore:
        subject_data['t2w'] = []

    anat_only = config.workflow.anat_only
    # Make sure we always go through these two checks
    if not anat_only and not subject_data['bold']:
        task_id = config.execution.task_id
        raise RuntimeError(
            'No BOLD images found for participant {} and task {}. '
            'All workflows require BOLD images.'.format(
                subject_id, task_id if task_id else '<all>'
            )
        )

    bold_runs = [
        sorted(
            listify(run),
            key=lambda fl: config.execution.layout.get_metadata(fl).get('EchoTime', 0),
        )
        for run in subject_data['bold']
    ]

    if subject_data['roi']:
        warnings.warn(
            f'Lesion mask {subject_data["roi"]} found. '
            'Future versions of NiBabies will use alternative conventions. '
            'Please refer to the documentation before upgrading.',
            FutureWarning,
            stacklevel=1,
        )

    recon_method = config.workflow.surface_recon_method
    msm_sulc = False

    anatomical_cache = {}
    if config.execution.derivatives:
        from nibabies.utils.derivatives import collect_anatomical_derivatives

        std_spaces = spaces.get_spaces(nonstandard=False, dim=(3,))
        std_spaces.append('fsnative')
        for deriv_dir in config.execution.derivatives.values():
            anatomical_cache.update(
                collect_anatomical_derivatives(
                    derivatives_dir=deriv_dir,
                    subject_id=subject_id,
                    session_id=session_id,
                    std_spaces=std_spaces,
                )
            )

        if config.execution.copy_derivatives:
            from nibabies.utils.derivatives import copy_derivatives

            config.loggers.workflow.info('Copying found anat derivatives into output directory')
            copy_derivatives(
                derivs=anatomical_cache,
                outdir=config.execution.nibabies_dir,
                modality='anat',
                subject_id=f'sub-{subject_id}',
                session_id=f'ses-{session_id}' if session_id else None,
            )

    # Determine some session level options here, as we should have
    # all the required information
    if recon_method == 'auto':
        if age <= MCRIBS_RECOMMEND_AGE_CAP and anatomical_cache.get('t2w_aseg'):
            # do not force mcribs without a vetted segmentation
            recon_method = 'mcribs'
        elif age <= 24:
            recon_method = 'infantfs'
        else:
            recon_method = 'freesurfer'

    requested_anat = config.execution.reference_anat
    t1w = subject_data['t1w']
    t2w = subject_data['t2w']
    single_anat = False

    if not (t1w and t2w):
        single_anat = True
        reference_anat = 'T1w' if t1w else 'T2w'
        if requested_anat and reference_anat != requested_anat:
            raise AttributeError(
                f'Requested to use {requested_anat} as anatomical reference but none available'
            )
    elif (reference_anat := requested_anat) is None:  # Both available with no preference
        reference_anat = (
            'T2w'
            if any(
                (
                    recon_method is None and age <= MCRIBS_RECOMMEND_AGE_CAP,
                    recon_method == 'mcribs',
                )
            )
            else 'T1w'
        )

    anat = reference_anat.lower()  # To be used for workflow connections

    config.loggers.workflow.info(
        'Collected the following data for %s:\nRaw:\n%s\n\nDerivatives:\n\n%s\n',
        f'sub-{subject_id}' if not session_id else f'sub-{subject_id}_ses-{session_id}',
        pprint.pformat(subject_data),
        pprint.pformat(anatomical_cache),
    )

    bids_root = str(config.execution.bids_dir)
    output_dir = str(config.execution.nibabies_dir)
    omp_nthreads = config.nipype.omp_nthreads

    inputnode = pe.Node(niu.IdentityInterface(fields=['subjects_dir']), name='inputnode')

    bidssrc = pe.Node(
        BIDSDataGrabber(
            subject_data=subject_data,
            anat_only=config.workflow.anat_only,
            anat_derivatives=anatomical_cache or None,
            subject_id=subject_id,
            require_t1w=False,
        ),
        name='bidssrc',
    )

    bids_info = pe.Node(
        BIDSInfo(bids_dir=config.execution.bids_dir, bids_validate=False),
        name='bids_info',
    )

    summary = pe.Node(
        SubjectSummary(
            anatomical_reference=reference_anat,
            recon_method=recon_method,
            std_spaces=spaces.get_spaces(nonstandard=False),
            nstd_spaces=spaces.get_spaces(standard=False),
            age=age,
        ),
        name='summary',
        run_without_submitting=True,
    )

    about = pe.Node(
        AboutSummary(version=config.environment.version, command=' '.join(sys.argv)),
        name='about',
        run_without_submitting=True,
    )

    ds_report_summary = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='summary',
            datatype='figures',
            dismiss_entities=('echo',),
        ),
        name='ds_report_summary',
        run_without_submitting=True,
    )

    ds_report_about = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='about',
            datatype='figures',
            dismiss_entities=('echo',),
        ),
        name='ds_report_about',
        run_without_submitting=True,
    )

    output_dir = config.execution.nibabies_dir
    sloppy = config.execution.sloppy
    cifti_output = config.workflow.cifti_output

    wf_args = {
        'age_months': age,
        't1w': t1w,
        't2w': t2w,
        'flair': subject_data['flair'],
        'bids_root': bids_root,
        'longitudinal': config.workflow.longitudinal,
        'msm_sulc': msm_sulc,
        'omp_nthreads': omp_nthreads,
        'output_dir': output_dir,
        'precomputed': anatomical_cache,
        'segmentation_atlases': config.execution.segmentation_atlases_dir,
        'skull_strip_fixed_seed': config.workflow.skull_strip_fixed_seed,
        'skull_strip_mode': config.workflow.skull_strip_anat,
        'skull_strip_template': Reference.from_string(config.workflow.skull_strip_template)[0],
        'recon_method': recon_method,
        'reference_anat': reference_anat,
        'sloppy': sloppy,
        'spaces': spaces,
        'cifti_output': cifti_output,
    }

    fit_wf = init_infant_single_anat_fit_wf if single_anat else init_infant_anat_fit_wf
    anat_fit_wf = fit_wf(**wf_args)

    # allow to run with anat-fast-track on fMRI-only dataset
    if f'{anat}_preproc' in anatomical_cache and not subject_data[anat]:
        workflow.connect([
            (bidssrc, bids_info, [(('bold', fix_multi_source_name), 'in_file')]),
            (anat_fit_wf, summary, [('outputnode.anat_preproc', anat)]),
            (anat_fit_wf, ds_report_summary, [('outputnode.anat_preproc', 'source_file')]),
            (anat_fit_wf, ds_report_about, [('outputnode.anat_preproc', 'source_file')]),
        ])  # fmt:skip
    else:
        workflow.connect([
            (bidssrc, bids_info, [((anat, fix_multi_source_name), 'in_file')]),
            (bidssrc, summary, [('t1w', 't1w')]),
            (bidssrc, ds_report_summary, [((anat, fix_multi_source_name), 'source_file')]),
            (bidssrc, ds_report_about, [((anat, fix_multi_source_name), 'source_file')]),
        ])  # fmt:skip

    if single_anat:
        workflow.connect(bidssrc, anat, anat_fit_wf, 'inputnode.anat')

    workflow.connect([
        (inputnode, anat_fit_wf, [('subjects_dir', 'inputnode.subjects_dir')]),
        (bidssrc, anat_fit_wf, [
            ('t1w', 'inputnode.t1w'),
            ('t2w', 'inputnode.t2w'),
            ('roi', 'inputnode.roi'),
            ('flair', 'inputnode.flair'),
        ]),
        # Reporting connections
        (inputnode, summary, [('subjects_dir', 'subjects_dir')]),
        (bidssrc, summary, [('t2w', 't2w'), ('bold', 'bold')]),
        (bids_info, summary, [
            ('subject', 'subject_id'),
            ('session', 'session_id'),
        ]),
        (summary, ds_report_summary, [('out_report', 'in_file')]),
        (summary, anat_fit_wf, [('subject_id', 'inputnode.subject_id')]),
        (about, ds_report_about, [('out_report', 'in_file')]),
    ])  # fmt:skip

    reg_sphere = f'sphere_reg_{"msm" if msm_sulc else "fsLR"}'
    template_iterator_wf = None
    select_MNIInfant_xfm = None
    if config.workflow.level == 'full':
        anat_apply_wf = init_infant_anat_apply_wf(
            bids_root=bids_root,
            cifti_output=cifti_output,
            msm_sulc=msm_sulc,
            omp_nthreads=omp_nthreads,
            output_dir=output_dir,
            precomputed=anatomical_cache,
            recon_method=recon_method,
            reference_anat=reference_anat,
            spaces=spaces,
        )
        workflow.connect([
            (anat_fit_wf, anat_apply_wf, [
                ('outputnode.anat_valid_list', 'inputnode.anat_valid_list'),
                ('outputnode.anat_preproc', 'inputnode.anat_preproc'),
                ('outputnode.anat_mask', 'inputnode.anat_mask'),
                ('outputnode.anat_dseg', 'inputnode.anat_dseg'),
                ('outputnode.anat_tpms', 'inputnode.anat_tpms'),
                ('outputnode.fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
                ('outputnode.white', 'inputnode.white'),
                ('outputnode.pial', 'inputnode.pial'),
                ('outputnode.midthickness', 'inputnode.midthickness'),
                (f'outputnode.{reg_sphere}', f'inputnode.{reg_sphere}'),
                ('outputnode.sulc', 'inputnode.sulc'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.thickness', 'inputnode.thickness'),
            ]),
        ])  # fmt:skip

        if cifti_output and 'MNIInfant' in [ref.space for ref in spaces.references]:
            mniinfant_res = 2 if config.workflow.cifti_output == '91k' else 1

            select_MNIInfant_xfm = pe.Node(
                KeySelect(
                    fields=['anat2std_xfm', 'std2anat_xfm'],
                    key=get_MNIInfant_key(spaces, mniinfant_res),
                ),
                name='select_MNIInfant_xfm',
                run_without_submitting=True,
            )

            workflow.connect([
                (anat_fit_wf, select_MNIInfant_xfm, [
                    ('outputnode.std2anat_xfm', 'std2anat_xfm'),
                    ('outputnode.anat2std_xfm', 'anat2std_xfm'),
                    ('outputnode.template', 'keys'),
                ]),
            ])  # fmt:skip

        if spaces.cached.get_spaces(nonstandard=False, dim=(3,)):
            template_iterator_wf = init_template_iterator_wf(spaces=spaces, sloppy=sloppy)

            workflow.connect([
                (anat_fit_wf, template_iterator_wf, [
                    ('outputnode.template', 'inputnode.template'),
                    ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                ]),
                (template_iterator_wf, anat_apply_wf, [
                    ('outputnode.std_t1w', 'inputnode.std_t1w',),
                    ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                    ('outputnode.space', 'inputnode.std_space'),
                    ('outputnode.cohort', 'inputnode.std_cohort'),
                    ('outputnode.resolution', 'inputnode.std_resolution'),
                ]),
            ])  # fmt:skip

    if config.workflow.anat_only:
        return clean_datasinks(workflow)

    fmap_estimators, estimator_map = map_fieldmap_estimation(
        layout=config.execution.layout,
        subject_id=subject_id,
        bold_data=bold_runs,
        ignore_fieldmaps='fieldmaps' in config.workflow.ignore,
        use_syn=config.workflow.use_syn_sdc,
        force_syn=config.workflow.force_syn,
        filters=config.execution.get().get('bids_filters', {}).get('fmap'),
    )

    if fmap_estimators:
        config.loggers.workflow.info(
            'B0 field inhomogeneity map will be estimated with the following '
            f'{len(fmap_estimators)} estimator(s): '
            f'{[e.method for e in fmap_estimators]}.'
        )

        from sdcflows import fieldmaps as fm
        from sdcflows.workflows.base import init_fmap_preproc_wf

        fmap_wf = init_fmap_preproc_wf(
            debug='fieldmaps' in config.execution.debug,
            estimators=fmap_estimators,
            omp_nthreads=omp_nthreads,
            output_dir=output_dir,
            subject=subject_id,
            sd_prior=False,  # No priors for infants yet
        )
        fmap_wf.__desc__ = f"""

Preprocessing of B<sub>0</sub> inhomogeneity mappings

: A total of {len(fmap_estimators)} fieldmaps were found available within the input
BIDS structure for this particular subject.
"""

        # Overwrite ``out_path_base`` of sdcflows's DataSinks
        for node in fmap_wf.list_node_names():
            if node.split('.')[-1].startswith('ds_'):
                fmap_wf.get_node(node).interface.out_path_base = ''

        # MG: No prior is used ATM, so no need for xfm
        # fmap_select_std = pe.Node(
        #     KeySelect(fields=['std2anat_xfm'], key='MNI152NLin2009cAsym'),
        #     name='fmap_select_std',
        #     run_without_submitting=True,
        # )
        # if any(estimator.method == fm.EstimatorType.ANAT for estimator in fmap_estimators):
        #     workflow.connect([
        #         (anat_fit_wf, fmap_select_std, [
        #             ('outputnode.std2anat_xfm', 'std2anat_xfm'),
        #             ('outputnode.template', 'keys')]),
        #     ])  # fmt:skip

        for estimator in fmap_estimators:
            config.loggers.workflow.info(
                f"""\
Setting-up fieldmap "{estimator.bids_id}" ({estimator.method}) with \
<{', '.join(s.path.name for s in estimator.sources)}>"""
            )

            # Mapped and phasediff can be connected internally by SDCFlows
            if estimator.method in (fm.EstimatorType.MAPPED, fm.EstimatorType.PHASEDIFF):
                continue

            suffices = [s.suffix for s in estimator.sources]

            if estimator.method == fm.EstimatorType.PEPOLAR:
                if len(suffices) == 2 and all(suf in ('epi', 'bold', 'sbref') for suf in suffices):
                    wf_inputs = getattr(fmap_wf.inputs, f'in_{estimator.bids_id}')
                    wf_inputs.in_data = [str(s.path) for s in estimator.sources]
                    wf_inputs.metadata = [s.metadata for s in estimator.sources]
                else:
                    raise NotImplementedError('Sophisticated PEPOLAR schemes are unsupported.')

            elif estimator.method == fm.EstimatorType.ANAT:
                from sdcflows.workflows.fit.syn import init_syn_preprocessing_wf

                sources = [str(s.path) for s in estimator.sources if s.suffix in ('bold', 'sbref')]
                source_meta = [
                    s.metadata for s in estimator.sources if s.suffix in ('bold', 'sbref')
                ]
                syn_preprocessing_wf = init_syn_preprocessing_wf(
                    omp_nthreads=omp_nthreads,
                    debug=config.execution.sloppy,
                    auto_bold_nss=True,
                    t1w_inversion=False,
                    sd_prior=False,
                    name=f'syn_preprocessing_{estimator.bids_id}',
                )
                syn_preprocessing_wf.inputs.inputnode.in_epis = sources
                syn_preprocessing_wf.inputs.inputnode.in_meta = source_meta

                workflow.connect([
                    (anat_fit_wf, syn_preprocessing_wf, [
                        ('outputnode.anat_preproc', 'inputnode.in_anat'),
                        ('outputnode.anat_mask', 'inputnode.mask_anat'),
                    ]),
                    # MG: No prior is used ATM, so no need for xfm
                    # (fmap_select_std, syn_preprocessing_wf, [
                    #     ('std2anat_xfm', 'inputnode.std2anat_xfm'),
                    # ]),
                    (syn_preprocessing_wf, fmap_wf, [
                        ('outputnode.epi_ref', f'in_{estimator.bids_id}.epi_ref'),
                        ('outputnode.epi_mask', f'in_{estimator.bids_id}.epi_mask'),
                        ('outputnode.anat_ref', f'in_{estimator.bids_id}.anat_ref'),
                        ('outputnode.anat_mask', f'in_{estimator.bids_id}.anat_mask'),
                        ('outputnode.sd_prior', f'in_{estimator.bids_id}.sd_prior'),
                    ]),
                ])  # fmt:skip

    # Append the functional section to the existing anatomical excerpt
    # That way we do not need to stream down the number of bold datasets
    func_pre_desc = f"""
Functional data preprocessing

: For each of the {len(bold_runs)} BOLD runs found per subject (across all
tasks and sessions), the following preprocessing was performed.
"""

    # Before initializing BOLD workflow, select/verify anatomical target for coregistration
    if config.workflow.bold2anat_init in ('auto', 't2w'):
        has_t2w = subject_data['t2w'] or 't2w_preproc' in anatomical_cache
        if config.workflow.bold2anat_init == 't2w' and not has_t2w:
            raise OSError(
                'A T2w image is expected for BOLD-to-anatomical coregistration and was not found'
            )
        config.workflow.bold2anat_init = 't2w' if has_t2w else 't1w'

    for bold_series in bold_runs:
        bold_file = bold_series[0]
        fieldmap_id = estimator_map.get(bold_file)

        functional_cache = {}
        if config.execution.derivatives:
            from nibabies.utils.bids import extract_entities
            from nibabies.utils.derivatives import collect_functional_derivatives

            entities = extract_entities(bold_series)

            for deriv_dir in config.execution.derivatives.values():
                functional_cache.update(
                    collect_functional_derivatives(
                        derivatives_dir=deriv_dir,
                        entities=entities,
                        fieldmap_id=fieldmap_id,
                    )
                )

            if config.execution.copy_derivatives:
                from nibabies.utils.derivatives import copy_derivatives

                config.loggers.workflow.info(
                    'Copying found func derivatives into output directory'
                )
                copy_derivatives(
                    derivs=functional_cache,
                    outdir=config.execution.nibabies_dir,
                    modality='func',
                    subject_id=f'sub-{subject_id}',
                    session_id=f'ses-{session_id}' if session_id else None,
                )

        bold_wf = init_bold_wf(
            reference_anat=reference_anat,
            bold_series=bold_series,
            precomputed=functional_cache,
            fieldmap_id=fieldmap_id,
            spaces=spaces,
        )
        if bold_wf is None:
            continue

        bold_wf.__desc__ = func_pre_desc + (bold_wf.__desc__ or '')

        workflow.connect([
            (anat_fit_wf, bold_wf, [
                ('outputnode.anat_preproc', 'inputnode.anat_preproc'),
                ('outputnode.anat_mask', 'inputnode.anat_mask'),
                ('outputnode.anat_dseg', 'inputnode.anat_dseg'),
                ('outputnode.anat_tpms', 'inputnode.anat_tpms'),
                ('outputnode.subjects_dir', 'inputnode.subjects_dir'),
                ('outputnode.subject_id', 'inputnode.subject_id'),
                ('outputnode.fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
                ('outputnode.white', 'inputnode.white'),
                ('outputnode.pial', 'inputnode.pial'),
                ('outputnode.midthickness', 'inputnode.midthickness'),
                ('outputnode.anat_ribbon', 'inputnode.anat_ribbon'),
                (f'outputnode.{reg_sphere}', 'inputnode.sphere_reg_fsLR'),
            ]),
        ])  # fmt:skip
        if fieldmap_id:
            workflow.connect([
                (fmap_wf, bold_wf, [
                    ('outputnode.fmap', 'inputnode.fmap'),
                    ('outputnode.fmap_ref', 'inputnode.fmap_ref'),
                    ('outputnode.fmap_coeff', 'inputnode.fmap_coeff'),
                    ('outputnode.fmap_mask', 'inputnode.fmap_mask'),
                    ('outputnode.fmap_id', 'inputnode.fmap_id'),
                    ('outputnode.method', 'inputnode.sdc_method'),
                ]),
            ])  # fmt:skip

        if config.workflow.level == 'full':
            if template_iterator_wf is not None:
                workflow.connect([
                    (template_iterator_wf, bold_wf, [
                        ('outputnode.space', 'inputnode.std_space'),
                        ('outputnode.resolution', 'inputnode.std_resolution'),
                        ('outputnode.cohort', 'inputnode.std_cohort'),
                        ('outputnode.std_t1w', 'inputnode.std_t1w'),
                        ('outputnode.std_mask', 'inputnode.std_mask'),
                        ('outputnode.anat2std_xfm', 'inputnode.anat2std_xfm'),
                    ]),
                ])  # fmt:skip

            if select_MNIInfant_xfm is not None:
                workflow.connect([
                    (select_MNIInfant_xfm, bold_wf, [
                        ('std2anat_xfm', 'inputnode.mniinfant2anat_xfm'),
                        ('anat2std_xfm', 'inputnode.anat2mniinfant_xfm')
                    ]),
                ])  # fmt:skip

            if config.workflow.cifti_output:
                workflow.connect([
                    (anat_apply_wf, bold_wf, [
                        ('outputnode.roi', 'inputnode.cortex_mask'),
                        ('outputnode.midthickness_fsLR', 'inputnode.midthickness_fsLR'),
                        ('outputnode.anat_aseg', 'inputnode.anat_aseg'),
                    ]),
                ])  # fmt:skip

    return clean_datasinks(workflow)


def _subject_session_id(subject_id: str, session_id: str | None) -> str:
    """
    Combine a subject ID with a session ID (if available).

    >>> _subject_session_id('01', None)
    'sub-01'
    >>> _subject_session_id('sub-01', '01')
    'sub-01_ses-01'
    >>> _subject_session_id('01', 'ses-03')
    'sub-01_ses-03'
    """
    entities = []
    entities.append(f'sub-{subject_id}' if not subject_id.startswith('sub-') else subject_id)
    if session_id is not None:
        entities.append(f'ses-{session_id}' if not session_id.startswith('ses-') else session_id)
    return '_'.join(entities)


def clean_datasinks(workflow: pe.Workflow) -> pe.Workflow:
    # Overwrite ``out_path_base`` of smriprep's DataSinks
    for node in workflow.list_node_names():
        if node.split('.')[-1].startswith('ds_'):
            workflow.get_node(node).interface.out_path_base = ''
    return workflow


def init_workflow_spaces(execution_spaces: SpatialReferences, age_months: int):
    """
    Create output spaces at a per-subworkflow level.

    This address the case where a multi-session subject is run,
    and requires separate template cohorts.
    """
    from niworkflows.utils.spaces import Reference

    from nibabies.utils.misc import cohort_by_months

    spaces = deepcopy(execution_spaces)

    if age_months is None:
        raise RuntimeError('Participant age (in months) is required.')

    if not spaces.references:
        # Ensure age specific template is added if nothing is present
        cohort = cohort_by_months('MNIInfant', age_months)
        spaces.add(('MNIInfant', {'res': 'native', 'cohort': cohort}))

    if not spaces.is_cached():
        spaces.checkpoint()

    # Ensure one cohort of MNIInfant is always available as an internal space
    if not any(
        space.startswith('MNIInfant') for space in spaces.get_spaces(nonstandard=False, dim=(3,))
    ):
        cohort = cohort_by_months('MNIInfant', age_months)
        spaces.add(Reference('MNIInfant', {'cohort': cohort}))

    if config.workflow.cifti_output:
        # CIFTI grayordinates to corresponding FSL-MNI resolutions.
        vol_res = '2' if config.workflow.cifti_output == '91k' else '1'
        spaces.add(Reference('MNI152NLin6Asym', {'res': vol_res}))
        # Ensure a non-native version of MNIInfant is added as a target
        cohort = cohort_by_months('MNIInfant', age_months)
        spaces.add(Reference('MNIInfant', {'cohort': cohort, 'res': vol_res}))

    return spaces


def init_execution_spaces():
    """Initialize the spaces to be saved.

    Either invoked by ``--output-spaces``,
    or an empty :py:class:`~niworkflows.utils.spaces.SpatialReferences`
    """
    from niworkflows.utils.spaces import Reference, SpatialReferences

    spaces = config.execution.output_spaces or SpatialReferences()
    if not isinstance(spaces, SpatialReferences):
        spaces = SpatialReferences(
            [ref for s in spaces.split(' ') for ref in Reference.from_string(s)]
        )
    return spaces


def map_fieldmap_estimation(
    layout: BIDSLayout,
    subject_id: str,
    bold_data: list[list[str]],
    ignore_fieldmaps: bool,
    use_syn: bool | str,
    force_syn: bool,
    filters: dict | None,
) -> tuple[list, dict]:
    if not any((not ignore_fieldmaps, use_syn, force_syn)):
        return [], {}

    from sdcflows import fieldmaps as fm
    from sdcflows.utils.wrangler import find_estimators

    # In the case where fieldmaps are ignored and `--use-syn-sdc` is requested,
    # SDCFlows `find_estimators` still receives a full layout (which includes the fmap modality)
    # and will not calculate fmapless schemes.
    # Similarly, if fieldmaps are ignored and `--force-syn` is requested,
    # `fmapless` should be set to True to ensure BOLD targets are found to be corrected.
    fmap_estimators = find_estimators(
        layout=layout,
        subject=subject_id,
        fmapless=bool(use_syn) or ignore_fieldmaps and force_syn,
        force_fmapless=force_syn or ignore_fieldmaps and use_syn,
        bids_filters=filters,
        anat_suffix=['T1w', 'T2w'],
    )

    if not fmap_estimators:
        if use_syn:
            message = (
                'Fieldmap-less (SyN) estimation was requested, but PhaseEncodingDirection '
                'information appears to be absent.'
            )
            config.loggers.workflow.error(message)
            if use_syn == 'error':
                raise ValueError(message)
        return [], {}

    if ignore_fieldmaps:
        if any(f.method == fm.EstimatorType.ANAT for f in fmap_estimators):
            config.loggers.workflow.info(
                'Option "--ignore fieldmaps" was set, but either "--use-syn-sdc" '
                'or "--force-syn" were given, so fieldmap-less estimation will be executed.'
            )
            fmap_estimators = [f for f in fmap_estimators if f.method == fm.EstimatorType.ANAT]
        else:
            config.loggers.workflow.info('Ignoring fieldmaps - no estimators will be used.')
            return [], {}

    # Pare down estimators to those that are actually used
    # If fmap_estimators == [], all loops/comprehensions terminate immediately
    all_ids = {fmap.bids_id for fmap in fmap_estimators}
    bold_files = (bold_series[0] for bold_series in bold_data)

    all_estimators = {
        bold_file: [fmap_id for fmap_id in get_estimator(layout, bold_file) if fmap_id in all_ids]
        for bold_file in bold_files
    }

    for bold_file, estimator_key in all_estimators.items():
        if len(estimator_key) > 1:
            config.loggers.workflow.warning(
                f'Several fieldmaps <{", ".join(estimator_key)}> are '
                f"'IntendedFor' <{bold_file}>, using {estimator_key[0]}"
            )
            estimator_key[1:] = []

    # Final, 1-1 map, dropping uncorrected BOLD
    estimator_map = {
        bold_file: estimator_key[0]
        for bold_file, estimator_key in all_estimators.items()
        if estimator_key
    }

    fmap_estimators = [f for f in fmap_estimators if f.bids_id in estimator_map.values()]

    return fmap_estimators, estimator_map


def get_estimator(layout, fname):
    field_source = layout.get_metadata(fname).get('B0FieldSource')
    if isinstance(field_source, str):
        field_source = (field_source,)

    if field_source is None:
        import re
        from pathlib import Path

        from sdcflows.fieldmaps import get_identifier

        # Fallback to IntendedFor
        intended_rel = re.sub(r'^sub-[a-zA-Z0-9]*/', '', str(Path(fname).relative_to(layout.root)))
        field_source = get_identifier(intended_rel)

    return field_source


def get_MNIInfant_key(spaces: SpatialReferences, res: str | int) -> str:
    """Parse spaces and return matching MNIInfant space, including cohort."""
    for ref in spaces.references:
        if ref.space == 'MNIInfant' and f'res-{res}' in str(ref):
            return ref.fullname

    raise KeyError(f'MNIInfant (resolution {res}) not found in SpatialReferences: {spaces}')



"""Anatomical surface projections"""

import typing as ty

import templateflow.api as tf
from nipype.interfaces import freesurfer as fs
from nipype.interfaces import io as nio
from nipype.interfaces import utility as niu
from nipype.interfaces.ants import N4BiasFieldCorrection
from nipype.pipeline import engine as pe
from niworkflows.engine import Workflow, tag
from niworkflows.interfaces.freesurfer import (
    PatchedLTAConvert as LTAConvert,
)
from niworkflows.interfaces.freesurfer import (
    PatchedRobustRegister as RobustRegister,
)
from niworkflows.interfaces.morphology import BinaryDilation
from niworkflows.interfaces.patches import FreeSurferSource
from smriprep.interfaces.freesurfer import MakeMidthickness
from smriprep.interfaces.workbench import SurfaceResample
from smriprep.workflows.surfaces import _extract_fs_fields

SURFACE_INPUTS = [
    't1w',
    't2w',
    'flair',
    'skullstripped_t1',
    'subjects_dir',
    'subject_id',
    # Customize aseg
    'in_aseg',
    'in_mask',
]
SURFACE_OUTPUTS = [
    'subjects_dir',
    'subject_id',
    'anat2fsnative_xfm',
    'fsnative2anat_xfm',
]


@tag('anat.recon')
def init_mcribs_surface_recon_wf(
    *,
    omp_nthreads: int,
    use_aseg: bool,
    precomputed: dict,
    mcribs_dir: str | None = None,
    name: str = 'mcribs_surface_recon_wf',
):
    """
    Reconstruct cortical surfaces using the M-CRIB-S pipeline.

    This workflow injects a precomputed segmentation into the M-CRIB-S pipeline, bypassing the
    DrawEM segmentation step that is normally performed.
    """
    from niworkflows.interfaces.nibabel import MapLabels, ReorientImage

    from nibabies.interfaces.mcribs import MCRIBReconAll

    if not use_aseg:
        raise NotImplementedError(
            'A previously computed segmentation is required for the M-CRIB-S workflow.'
        )

    inputnode = pe.Node(niu.IdentityInterface(fields=SURFACE_INPUTS), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=SURFACE_OUTPUTS), name='outputnode')

    workflow = Workflow(name=name)
    workflow.__desc__ = (
        'Brain surfaces were reconstructed with a modified `MCRIBReconAll` [M-CRIB-S, @mcribs]'
        'workflow, using the reference T2w and a pre-computed anatomical segmentation'
    )

    # mapping of labels from FS to M-CRIB-S (DrawEM)
    fs2mcribs = {
        1: 21,
        2: 51,
        3: 21,
        4: 49,
        5: 0,
        6: 17,
        7: 17,
        8: 17,
        10: 43,
        11: 41,
        12: 47,
        13: 47,
        14: 114,
        15: 0,
        16: 19,
        17: 1,
        18: 3,
        24: 83,
        26: 41,
        28: 45,
        30: 51,
        31: 49,
        41: 52,
        42: 20,
        43: 50,
        44: 0,
        46: 18,
        47: 18,
        49: 42,
        50: 40,
        51: 46,
        52: 46,
        53: 2,
        54: 4,
        58: 40,
        60: 44,
        62: 25,
        63: 50,
        77: 51,
        85: 21,
        172: 172,
        253: 48,
    }
    fs_to_mcribs = pe.Node(MapLabels(mappings=fs2mcribs), name='fs_to_mcribs')

    t2w_las = pe.Node(ReorientImage(target_orientation='LAS'), name='t2w_las')
    seg_las = pe.Node(ReorientImage(target_orientation='LAS'), name='seg_las')

    # dilated mask and use in recon-neonatal-cortex
    mask_dil = pe.Node(BinaryDilation(radius=3), name='mask_dil')
    mask_las = pe.Node(ReorientImage(target_orientation='LAS'), name='mask_las')

    # N4BiasCorrection occurs in MCRIBTissueSegMCRIBS (which is skipped)
    # Run it (with mask to rescale intensities) prior injection
    n4_mcribs = pe.Node(
        N4BiasFieldCorrection(
            dimension=3,
            bspline_fitting_distance=200,
            save_bias=True,
            copy_header=True,
            n_iterations=[50] * 5,
            convergence_threshold=1e-7,
            rescale_intensities=True,
            shrink_factor=4,
        ),
        name='n4_mcribs',
    )

    mcribs_recon = pe.Node(
        MCRIBReconAll(
            surfrecon=True,
            surfrecon_method='Deformable',
            join_thresh=1.0,
            fast_collision=True,
            nthreads=omp_nthreads,
            outdir=mcribs_dir,
        ),
        name='mcribs_recon',
        mem_gb=5,
    )
    mcribs_recon.config = {'execution': {'remove_unnecessary_outputs': False}}

    mcribs_postrecon = pe.Node(
        MCRIBReconAll(autorecon_after_surf=True, nthreads=omp_nthreads),
        name='mcribs_postrecon',
        mem_gb=5,
    )
    mcribs_postrecon.config = {'execution': {'remove_unnecessary_outputs': False}}

    fssource = pe.Node(FreeSurferSource(), name='fssource', run_without_submitting=True)
    midthickness_wf = init_make_midthickness_wf(omp_nthreads=omp_nthreads)

    workflow.connect([
        (inputnode, t2w_las, [('t2w', 'in_file')]),
        (inputnode, fs_to_mcribs, [('in_aseg', 'in_file')]),
        (inputnode, mask_dil, [('in_mask', 'in_mask')]),
        (mask_dil, mask_las, [('out_mask', 'in_file')]),
        (mask_las, mcribs_recon, [('out_file', 'mask_file')]),
        (fs_to_mcribs, seg_las, [('out_file', 'in_file')]),
        (inputnode, mcribs_recon, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id')]),
        (t2w_las, n4_mcribs, [('out_file', 'input_image')]),
        (mask_las, n4_mcribs, [('out_file', 'mask_image')]),
        (n4_mcribs, mcribs_recon, [('output_image', 't2w_file')]),
        (seg_las, mcribs_recon, [('out_file', 'segmentation_file')]),
        (inputnode, mcribs_postrecon, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id')]),
        (mcribs_recon, mcribs_postrecon, [('mcribs_dir', 'outdir')]),
        (mcribs_postrecon, fssource, [('subjects_dir', 'subjects_dir')]),
        (mcribs_postrecon, midthickness_wf, [('subjects_dir', 'inputnode.subjects_dir')]),
        (inputnode, fssource, [('subject_id', 'subject_id')]),
        (inputnode, midthickness_wf, [('subject_id', 'inputnode.subject_id')]),
        (fssource, midthickness_wf, [
            ('white', 'inputnode.white'),
            ('graymid', 'inputnode.graymid'),
        ]),
        (midthickness_wf, outputnode, [
            ('outputnode.subjects_dir', 'subjects_dir'),
            ('outputnode.subject_id', 'subject_id'),
        ]),
    ])  # fmt:skip

    if 'fsnative' not in precomputed.get('transforms', {}):
        fsnative2anat_xfm = pe.Node(
            RobustRegister(auto_sens=True, est_int_scale=True),
            name='fsnative2anat_xfm',
        )
        anat2fsnative_xfm = pe.Node(
            LTAConvert(out_lta=True, invert=True),
            name='anat2fsnative_xfm',
        )
        workflow.connect([
            (inputnode, fsnative2anat_xfm, [('t2w', 'target_file')]),
            (fssource, fsnative2anat_xfm, [('T2', 'source_file')]),
            (fsnative2anat_xfm, outputnode, [('out_reg_file', 'fsnative2anat_xfm')]),
            (fsnative2anat_xfm, anat2fsnative_xfm, [('out_reg_file', 'in_lta')]),
            (anat2fsnative_xfm, outputnode, [('out_lta', 'anat2fsnative_xfm')]),
        ])  # fmt:skip

    return workflow


@tag('anat.fslr-reg')
def init_mcribs_dhcp_wf(*, name='mcribs_dhcp_wf'):
    """
    Generate GIFTI registration files to dhcp (42-week) space.

    Note: The dhcp template was derived from the Conte69 atlas,
    and maps reasonably well to fsLR.
    """
    from smriprep.interfaces.workbench import SurfaceSphereProjectUnproject

    workflow = Workflow(name=name)

    inputnode = pe.Node(
        niu.IdentityInterface(['sphere_reg', 'sulc']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(['sphere_reg_dhcpAsym']),
        name='outputnode',
    )

    # SurfaceSphereProjectUnProject
    # project to 41k dHCP atlas sphere
    #   - sphere-in: Individual native sphere in surf directory registered to 41k atlas sphere
    #   - sphere-to: the 41k atlas sphere, in the fsaverage directory
    #   - sphere-unproject-from: 41k atlas sphere registered to dHCP 42wk sphere,
    #                            in the fsaverage directory
    #   - sphere-out: lh.sphere.reg2.dHCP42.native.surf.gii
    project_unproject = pe.MapNode(
        SurfaceSphereProjectUnproject(),
        iterfield=['sphere_in', 'sphere_project_to', 'sphere_unproject_from'],
        name='project_unproject',
    )
    project_unproject.inputs.sphere_project_to = [
        str(
            tf.get(
                'fsaverage',
                density='41k',
                hemi=hemi,
                desc=None,
                suffix='sphere',
                extension='.surf.gii',
            )
        )
        for hemi in 'LR'
    ]

    project_unproject.inputs.sphere_unproject_from = [  # TODO: Use symmetric template
        str(
            tf.get(
                'dhcpAsym',
                space='fsaverage',
                hemi=hemi,
                density='41k',
                desc='reg',
                suffix='sphere',
                extension='.surf.gii',
                raise_empty=True,
            )
        )
        for hemi in 'LR'
    ]

    workflow.connect([
        (inputnode, project_unproject, [('sphere_reg', 'sphere_in')]),
        (project_unproject, outputnode, [('sphere_out', 'sphere_reg_dhcpAsym')]),
    ])  # fmt:skip

    return workflow


@tag('anat.recon')
def init_infantfs_surface_recon_wf(
    *,
    age_months: int,
    precomputed: dict,
    omp_nthreads: int,
    use_aseg: bool = False,
    name: str = 'infantfs_surface_recon_wf',
):
    from nibabies.interfaces.freesurfer import InfantReconAll

    workflow = Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=SURFACE_INPUTS), name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(fields=SURFACE_OUTPUTS), name='outputnode')

    desc = (
        'Brain surfaces were reconstructed using `infant_recon_all` [FreeSurfer '
        f'{fs.Info().looseversion() or "<ver>"}, RRID:SCR_001847, @infantfs], '
        'using the reference T1w'
    )
    desc += '.' if not use_aseg else ' and a pre-computed anatomical segmentation.'
    workflow.__desc__ = desc

    gen_recon_outdir = pe.Node(niu.Function(function=_gen_recon_dir), name='gen_recon_outdir')

    # inject the intensity-normalized skull-stripped t1w from the brain extraction workflow
    recon = pe.Node(InfantReconAll(age=age_months), name='reconall')
    if use_aseg:
        workflow.connect(inputnode, 'in_aseg', recon, 'aseg_file')

    fssource = pe.Node(FreeSurferSource(), name='fssource', run_without_submitting=True)
    midthickness_wf = init_make_midthickness_wf(omp_nthreads=omp_nthreads)

    workflow.connect([
        (inputnode, gen_recon_outdir, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id'),
        ]),
        (inputnode, recon, [
            ('skullstripped_t1', 'mask_file'),
            ('subject_id', 'subject_id'),
        ]),
        (gen_recon_outdir, recon, [
            ('out', 'outdir'),
        ]),
        (recon, fssource, [
            ('subject_id', 'subject_id'),
            (('outdir', _parent), 'subjects_dir'),
        ]),
        (recon, midthickness_wf, [
            ('subject_id', 'inputnode.subject_id'),
            (('outdir', _parent), 'inputnode.subjects_dir'),
        ]),
        (fssource, midthickness_wf, [
            ('white', 'inputnode.white'),
            ('graymid', 'inputnode.graymid'),
        ]),
        (midthickness_wf, outputnode, [
            ('outputnode.subjects_dir', 'subjects_dir'),
            ('outputnode.subject_id', 'subject_id'),
        ])
    ])  # fmt:skip

    if 'fsnative' not in precomputed.get('transforms', {}):
        fsnative2anat_xfm = pe.Node(
            RobustRegister(auto_sens=True, est_int_scale=True),
            name='fsnative2anat_xfm',
        )
        anat2fsnative_xfm = pe.Node(
            LTAConvert(out_lta=True, invert=True),
            name='anat2fsnative_xfm',
        )
        workflow.connect([
            (inputnode, fsnative2anat_xfm, [('skullstripped_t1', 'target_file')]),
            (fssource, fsnative2anat_xfm, [
                (('norm', _replace_mgz), 'source_file'),
            ]),
            (fsnative2anat_xfm, anat2fsnative_xfm, [('out_reg_file', 'in_lta')]),
            (fsnative2anat_xfm, outputnode, [
                ('out_reg_file', 'fsnative2anat_xfm'),
            ]),
            (anat2fsnative_xfm, outputnode, [
                ('out_lta', 'anat2fsnative_xfm'),
            ]),
        ])  # fmt:skip

    return workflow


@tag('anat.midthickness')
def init_make_midthickness_wf(
    *, omp_nthreads: int, name: str = 'make_midthickness_wf'
) -> pe.Workflow:
    """
    Standalone workflow to create and save cortical midthickness, derived from
    the generated white / graymid surfaces.
    """

    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['subject_id', 'subjects_dir', 'white', 'graymid']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['subject_id', 'subjects_dir']),
        name='outputnode',
    )

    midthickness = pe.MapNode(
        MakeMidthickness(thickness=True, distance=0.5, out_name='midthickness'),
        iterfield='in_file',
        name='midthickness',
        n_procs=min(omp_nthreads, 12),
    )
    save_midthickness = pe.Node(nio.DataSink(parameterization=False), name='save_midthickness')

    sync = pe.Node(
        niu.Function(
            function=_extract_fs_fields,
            output_names=['subjects_dir', 'subject_id'],
        ),
        name='sync',
    )

    workflow.connect([
        (inputnode, midthickness, [
            ('white', 'in_file'),
            ('graymid', 'graymid'),
        ]),
        (midthickness, save_midthickness, [('out_file', 'surf.@graymid')]),
        (inputnode, save_midthickness, [
            ('subjects_dir', 'base_directory'),
            ('subject_id', 'container'),
        ]),
        (save_midthickness, sync, [('out_file', 'filenames')]),
        (sync, outputnode, [
            ('subjects_dir', 'subjects_dir'),
            ('subject_id', 'subject_id'),
        ]),
    ])  # fmt:skip
    return workflow


@tag('anat.resample-surfs')
def init_resample_surfaces_dhcp_wf(
    surfaces: list[str],
    grayord_density: ty.Literal['91k', '170k'],
    name: str = 'resample_surfaces_dhcp_wf',
):
    """
    Resample subject midthickness surface to specified density.

    Workflow Graph
        .. workflow::
            :graph2use: colored
            :simple_form: yes

            from nibabies.workflows.anatomical.surfaces import init_resample_surfaces_dhcp_wf
            wf = init_resample_surfaces_dhcp_wf(surfaces=['white', grayord_density='91k')

    Parameters
    ----------
    grayord_density : :obj:`str`
        Either `91k` or `170k`, representing the total of vertices or *grayordinates*.
    name : :obj:`str`
        Unique name for the subworkflow (default: ``"resample_surfaces_dhcp_wf``)

    Inputs
    ------
    ``<surface>``
        Left and right GIFTIs for each surface name passed to ``surfaces``
    sphere_reg_fsLR
        GIFTI surface mesh corresponding to the subject's fsLR registration sphere

    Outputs
    -------
    midthickness
        GIFTI surface mesh corresponding to the midthickness surface, resampled to fsLR
    """
    workflow = Workflow(name=name)

    fslr_density = '32k' if grayord_density == '91k' else '59k'

    inputnode = pe.Node(
        niu.IdentityInterface(fields=[*surfaces, 'sphere_reg_fsLR']),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=[f'{surf}_fsLR' for surf in surfaces]), name='outputnode'
    )

    surface_list = pe.Node(
        niu.Merge(len(surfaces), ravel_inputs=True),
        name='surface_list',
        run_without_submitting=True,
    )

    resampler = pe.MapNode(
        SurfaceResample(method='BARYCENTRIC'),
        iterfield=['surface_in', 'current_sphere', 'new_sphere'],
        name='resampler',
    )
    resampler.inputs.new_sphere = [
        str(
            tf.get(
                template='dhcpAsym',
                cohort='42',
                density=fslr_density,
                suffix='sphere',
                hemi=hemi,
                space=None,
                extension='.surf.gii',
            )
        )
        # Order matters. Iterate over surfaces, then hemis to get L R L R L R
        for _surf in surfaces
        for hemi in ['L', 'R']
    ]

    surface_groups = pe.Node(
        niu.Split(splits=[2] * len(surfaces)),
        name='surface_groups',
        run_without_submitting=True,
    )

    workflow.connect([
        (inputnode, surface_list, [
            ((surf, _sorted_by_basename), f'in{i}')
            for i, surf in enumerate(surfaces, start=1)
        ]),
        (inputnode, resampler, [
            (('sphere_reg_fsLR', _repeat, len(surfaces)), 'current_sphere'),
        ]),
        (surface_list, resampler, [('out', 'surface_in')]),
        (resampler, surface_groups, [('surface_out', 'inlist')]),
        (surface_groups, outputnode, [
            (f'out{i}', f'{surf}_fsLR') for i, surf in enumerate(surfaces, start=1)
        ]),
    ])  # fmt:skip

    return workflow


def _sorted_by_basename(inlist):
    from os.path import basename

    return sorted(inlist, key=lambda x: str(basename(x)))


def _repeat(seq: list, count: int) -> list:
    return seq * count


def _parent(p):
    from pathlib import Path

    return str(Path(p).parent)


def _gen_recon_dir(subjects_dir, subject_id):
    from pathlib import Path

    p = Path(subjects_dir) / subject_id
    p.mkdir(parents=True, exist_ok=True)
    return str(p)


def _replace_mgz(in_file):
    return in_file.replace('.mgz', '.nii.gz')



from pathlib import Path

from nipype.interfaces import (
    freesurfer as fs,
)
from nipype.interfaces.ants.base import ANTSCommand, ANTSCommandInputSpec
from nipype.interfaces.ants.registration import (
    CompositeTransformUtil as _CompositeTransformUtil,
)
from nipype.interfaces.ants.registration import (
    CompositeTransformUtilOutputSpec as _CompositeTransformUtilOutputSpec,
)
from nipype.interfaces.base import File, InputMultiObject, TraitedSpec, traits


class _MRICoregInputSpec(fs.registration.MRICoregInputSpec):
    reference_file = File(
        argstr='--ref %s',
        desc='reference (target) file',
        copyfile=False,
    )
    subject_id = traits.Str(
        argstr='--s %s',
        position=1,
        requires=['subjects_dir'],
        desc='freesurfer subject ID (implies ``reference_mask == '
        'aparc+aseg.mgz`` unless otherwise specified)',
    )


class MRICoreg(fs.MRICoreg):
    """
    Patched that allows setting both a reference file and the subjects dir.
    """

    input_spec = _MRICoregInputSpec


class ConcatXFMInputSpec(ANTSCommandInputSpec):
    transforms = InputMultiObject(
        traits.Either(File(exists=True), 'identity'),
        argstr='%s',
        mandatory=True,
        desc='transform files: will be applied in reverse order. For '
        'example, the last specified transform will be applied first.',
    )
    out_xfm = traits.File(
        'concat_xfm.h5',
        usedefault=True,
        argstr='--output [ %s, 1 ]',
        desc='output file name',
    )
    reference_image = File(
        argstr='--reference-image %s',
        mandatory=True,
        desc='reference image space that you wish to warp INTO',
        exists=True,
    )
    invert_transform_flags = InputMultiObject(traits.Bool())


class ConcatXFMOutputSpec(TraitedSpec):
    out_xfm = File(desc='Combined transform')


class ConcatXFM(ANTSCommand):
    """
    Streamed use of antsApplyTransforms to combine multiple xfms into a single file

    Examples
    --------

    >>> from nibabies.interfaces.patches import ConcatXFM
    >>> cxfm = ConcatXFM()
    >>> cxfm.inputs.transforms = [testdir / 'xfm0.h5', testdir / 'xfm1.h5']
    >>> cxfm.inputs.reference_image = testdir / 'anatomical.nii'
    >>> cxfm.cmdline  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
    'antsApplyTransforms --output [ concat_xfm.h5, 1 ] --reference-image .../anatomical.nii \
--transform .../xfm0.h5 --transform .../xfm1.h5'

    """

    _cmd = 'antsApplyTransforms'
    input_spec = ConcatXFMInputSpec
    output_spec = ConcatXFMOutputSpec

    def _get_transform_filenames(self):
        retval = []
        invert_flags = self.inputs.invert_transform_flags
        if not invert_flags:
            invert_flags = [False] * len(self.inputs.transforms)
        elif len(self.inputs.transforms) != len(invert_flags):
            raise ValueError(
                'ERROR: The invert_transform_flags list must have the same number '
                'of entries as the transforms list.'
            )

        for transform, invert in zip(self.inputs.transforms, invert_flags, strict=False):
            if invert:
                retval.append(f'--transform [ {transform}, 1 ]')
            else:
                retval.append(f'--transform {transform}')
        return ' '.join(retval)

    def _format_arg(self, opt, spec, val):
        if opt == 'transforms':
            return self._get_transform_filenames()
        return super()._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_xfm'] = Path(self.inputs.out_xfm).absolute()
        return outputs


class CompositeTransformUtilOutputSpec(_CompositeTransformUtilOutputSpec):
    out_transforms = traits.List(desc='list of ordered transform components')


class CompositeTransformUtil(_CompositeTransformUtil):
    """Outputs have changed in newer versions of ANTs."""

    output_spec = CompositeTransformUtilOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        # Ordering may change depending on forward/inverse transform
        # Forward: <prefix>_00_AffineTransform.mat, <prefix>_01_DisplacementFieldTransform.nii.gz
        # Inverse: <prefix>_01_AffineTransform.mat, <prefix>_00_DisplacementFieldTransform.nii.gz
        if self.inputs.process == 'disassemble':
            transforms = [
                str(Path(x).absolute())
                for x in sorted(Path().glob(f'{self.inputs.output_prefix}_*'))
            ]
            outputs['out_transforms'] = transforms

            # Potentially could be more than one affine / displacement per composite transform...
            outputs['affine_transform'] = [
                x for x in transforms if 'AffineTransform' in Path(x).name
            ][0]
            outputs['displacement_field'] = [
                x for x in transforms if 'DisplacementFieldTransform' in Path(x).name
            ][0]
        elif self.inputs.process == 'assemble':
            outputs['out_file'] = Path(self.inputs.out_file).absolute()
        return outputs



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Orchestrating the BOLD-preprocessing workflow
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_bold_wf
.. autofunction:: init_bold_fit_wf
.. autofunction:: init_bold_native_wf

"""

import typing as ty

from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.utils.connections import listify

from nibabies import config
from nibabies._types import Anatomical
from nibabies.interfaces import DerivativesDataSink
from nibabies.utils.misc import estimate_bold_mem_usage

# BOLD workflows
from nibabies.workflows.bold.apply import init_bold_volumetric_resample_wf
from nibabies.workflows.bold.confounds import init_bold_confs_wf, init_carpetplot_wf
from nibabies.workflows.bold.fit import init_bold_fit_wf, init_bold_native_wf
from nibabies.workflows.bold.outputs import (
    init_ds_bold_native_wf,
    init_ds_volumes_wf,
    prepare_timing_parameters,
)
from nibabies.workflows.bold.resampling import init_bold_surf_wf
from nibabies.workflows.bold.t2s import init_t2s_reporting_wf

if ty.TYPE_CHECKING:
    from niworkflows.utils.spaces import SpatialReferences

DEFAULT_DISMISS_ENTITIES = config.DEFAULT_DISMISS_ENTITIES


def init_bold_wf(
    *,
    reference_anat: Anatomical,
    bold_series: list[str],
    precomputed: dict | None = None,
    fieldmap_id: str | None = None,
    spaces: 'SpatialReferences',
) -> pe.Workflow:
    """
    This workflow controls the functional preprocessing stages of *fMRIPrep*.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.tests import mock_config
            from fmriprep import config
            from fmriprep.workflows.bold.base import init_bold_wf
            with mock_config():
                bold_file = config.execution.bids_dir / "sub-01" / "func" \
                    / "sub-01_task-mixedgamblestask_run-01_bold.nii.gz"
                wf = init_bold_wf(
                    bold_series=[str(bold_file)],
                )

    Parameters
    ----------
    bold_series
        List of paths to NIfTI files.
    precomputed
        Dictionary containing precomputed derivatives to reuse, if possible.
    fieldmap_id
        ID of the fieldmap to use to correct this BOLD series. If :obj:`None`,
        no correction will be applied.

    Inputs
    ------
    anat_preproc
        Bias-corrected structural template image
    anat_mask
        Mask of the skull-stripped template image
    anat_dseg
        Segmentation of preprocessed structural image, including
        gray-matter (GM), white-matter (WM) and cerebrospinal fluid (CSF)
    anat_tpms
        List of tissue probability maps in anatomical space
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    fsnative2anat_xfm
        LTA-style affine matrix translating from FreeSurfer-conformed subject space to anatomical
    white
        FreeSurfer white matter surfaces, in anat space, collated left, then right
    midthickness
        FreeSurfer mid-thickness surfaces, in anat space, collated left, then right
    pial
        FreeSurfer pial surfaces, in anat space, collated left, then right
    sphere_reg_fsLR
        Registration spheres from fsnative to fsLR space, collated left, then right
    anat_ribbon
        Binary cortical ribbon mask in anat space
    fmap_id
        Unique identifiers to select fieldmap files
    fmap
        List of estimated fieldmaps (collated with fmap_id)
    fmap_ref
        List of fieldmap reference files (collated with fmap_id)
    fmap_coeff
        List of lists of spline coefficient files (collated with fmap_id)
    fmap_mask
        List of fieldmap masks (collated with fmap_id)
    sdc_method
        List of fieldmap correction method names (collated with fmap_id)
    anat2std_xfm
        Transform from anatomical space to standard space
    std_t1w
        T1w reference image in standard space
    std_mask
        Brain (binary) mask of the standard reference image
    std_space
        Value of space entity to be used in standard space output filenames
    std_resolution
        Value of resolution entity to be used in standard space output filenames
    std_cohort
        Value of cohort entity to be used in standard space output filenames
    anat2mni6_xfm
        Transform from anatomical space to MNI152NLin6Asym space
    mni6_mask
        Brain (binary) mask of the MNI152NLin6Asym reference image
    mni2009c2anat_xfm
        Transform from MNI152NLin2009cAsym to anatomical space

    Note that ``anat2std_xfm``, ``std_space``, ``std_resolution``,
    ``std_cohort``, ``std_t1w`` and ``std_mask`` are treated as single
    inputs. In order to resample to multiple target spaces, connect
    these fields to an iterable.

    See Also
    --------

    * :func:`~fmriprep.workflows.bold.fit.init_bold_fit_wf`
    * :func:`~fmriprep.workflows.bold.fit.init_bold_native_wf`
    * :func:`~fmriprep.workflows.bold.apply.init_bold_volumetric_resample_wf`
    * :func:`~fmriprep.workflows.bold.outputs.init_ds_bold_native_wf`
    * :func:`~fmriprep.workflows.bold.outputs.init_ds_volumes_wf`
    * :func:`~fmriprep.workflows.bold.t2s.init_t2s_reporting_wf`
    * :func:`~fmriprep.workflows.bold.resampling.init_bold_surf_wf`
    * :func:`~fmriprep.workflows.bold.resampling.init_bold_fsLR_resampling_wf`
    * :func:`~fmriprep.workflows.bold.resampling.init_bold_grayords_wf`
    * :func:`~fmriprep.workflows.bold.confounds.init_bold_confs_wf`
    * :func:`~fmriprep.workflows.bold.confounds.init_carpetplot_wf`

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow

    if precomputed is None:
        precomputed = {}
    bold_file = bold_series[0]

    output_dir = str(config.execution.nibabies_dir)
    omp_nthreads = config.nipype.omp_nthreads
    all_metadata = [config.execution.layout.get_metadata(file) for file in bold_series]

    nvols, mem_gb = estimate_bold_mem_usage(bold_file)
    if nvols <= 5 - config.execution.sloppy:
        config.loggers.workflow.warning(
            f'Too short BOLD series (<= 5 timepoints). Skipping processing of <{bold_file}>.'
        )
        return

    config.loggers.workflow.debug(
        'Creating bold processing workflow for <%s> (%.2f GB / %d TRs). '
        'Memory resampled/largemem=%.2f/%.2f GB.',
        bold_file,
        mem_gb['filesize'],
        nvols,
        mem_gb['resampled'],
        mem_gb['largemem'],
    )

    workflow = Workflow(name=_get_wf_name(bold_file, 'bold'))
    workflow.__postdesc__ = """\
All resamplings can be performed with *a single interpolation
step* by composing all the pertinent transformations (i.e. head-motion
transform matrices, susceptibility distortion correction when available,
and co-registrations to anatomical and output spaces).
Gridded (volumetric) resamplings were performed using `nitransforms`,
configured with cubic B-spline interpolation.
"""

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                # Anatomical coregistration
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'anat_tpms',
                'anat_aseg',
                # FreeSurfer outputs
                'subjects_dir',
                'subject_id',
                'fsnative2anat_xfm',
                'white',
                'midthickness',
                'pial',
                'sphere_reg_fsLR',
                'midthickness_fsLR',
                'cortex_mask',
                'anat_ribbon',
                # Fieldmap registration
                'fmap',
                'fmap_ref',
                'fmap_coeff',
                'fmap_mask',
                'fmap_id',
                'sdc_method',
                # Volumetric templates
                'anat2std_xfm',
                'std_t1w',
                'std_mask',
                'std_space',
                'std_resolution',
                'std_cohort',
                # MNI152NLin6Asym warp, for CIFTI use
                # 'anat2mni6_xfm',
                # 'mni6_mask',
                # MNI152NLin2009cAsym inverse warp, for carpetplotting (NOT USED)
                # 'mni2009c2anat_xfm',
                # MNIInfant <cohort> warp, for CIFTI use
                'anat2mniinfant_xfm',
                'mniinfant2anat_xfm',
                'mniinfant_mask',
            ],
        ),
        name='inputnode',
    )

    #
    # Minimal workflow
    #

    bold_fit_wf = init_bold_fit_wf(
        bold_series=bold_series,
        reference_anat=reference_anat,
        precomputed=precomputed,
        fieldmap_id=fieldmap_id,
        omp_nthreads=omp_nthreads,
    )

    workflow.connect([
        (inputnode, bold_fit_wf, [
            ('anat_preproc', 'inputnode.anat_preproc'),
            ('anat_mask', 'inputnode.anat_mask'),
            ('anat_dseg', 'inputnode.anat_dseg'),
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
            ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ('fmap', 'inputnode.fmap'),
            ('fmap_ref', 'inputnode.fmap_ref'),
            ('fmap_coeff', 'inputnode.fmap_coeff'),
            ('fmap_mask', 'inputnode.fmap_mask'),
            ('fmap_id', 'inputnode.fmap_id'),
            ('sdc_method', 'inputnode.sdc_method'),
        ]),
    ])  # fmt:skip

    if config.workflow.level == 'minimal':
        return workflow

    # Now that we're resampling and combining, multiecho matters
    multiecho = len(bold_series) > 2

    nonstd_spaces = set(spaces.get_nonstandard())
    freesurfer_spaces = spaces.get_fs_spaces()

    #
    # Resampling outputs workflow:
    #   - Resample to native
    #   - Save native outputs/echos only if requested
    #

    bold_native_wf = init_bold_native_wf(
        bold_series=bold_series,
        fieldmap_id=fieldmap_id,
        omp_nthreads=omp_nthreads,
    )

    workflow.connect([
        (inputnode, bold_native_wf, [
            ('fmap_ref', 'inputnode.fmap_ref'),
            ('fmap_coeff', 'inputnode.fmap_coeff'),
            ('fmap_id', 'inputnode.fmap_id'),
        ]),
        (bold_fit_wf, bold_native_wf, [
            ('outputnode.coreg_boldref', 'inputnode.boldref'),
            ('outputnode.bold_mask', 'inputnode.bold_mask'),
            ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
            ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
            ('outputnode.dummy_scans', 'inputnode.dummy_scans'),
        ]),
    ])  # fmt:skip

    boldref_out = bool(nonstd_spaces.intersection(('func', 'run', 'bold', 'boldref', 'sbref')))
    boldref_out &= config.workflow.level == 'full'
    echos_out = multiecho and config.execution.me_output_echos

    if boldref_out or echos_out:
        ds_bold_native_wf = init_ds_bold_native_wf(
            bids_root=str(config.execution.bids_dir),
            output_dir=output_dir,
            bold_output=boldref_out,
            echo_output=echos_out,
            multiecho=multiecho,
            all_metadata=all_metadata,
        )
        ds_bold_native_wf.inputs.inputnode.source_files = bold_series

        workflow.connect([
            (bold_fit_wf, ds_bold_native_wf, [
                ('outputnode.bold_mask', 'inputnode.bold_mask'),
                ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
                ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
            ]),
            (bold_native_wf, ds_bold_native_wf, [
                ('outputnode.bold_native', 'inputnode.bold'),
                ('outputnode.bold_echos', 'inputnode.bold_echos'),
                ('outputnode.t2star_map', 'inputnode.t2star'),
            ]),
        ])  # fmt:skip

    if multiecho:
        t2s_reporting_wf = init_t2s_reporting_wf()

        ds_report_t2scomp = pe.Node(
            DerivativesDataSink(
                desc='t2scomp',
                datatype='figures',
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            ),
            name='ds_report_t2scomp',
            run_without_submitting=True,
        )

        ds_report_t2star_hist = pe.Node(
            DerivativesDataSink(
                desc='t2starhist',
                datatype='figures',
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
            ),
            name='ds_report_t2star_hist',
            run_without_submitting=True,
        )

        workflow.connect([
            (inputnode, t2s_reporting_wf, [('anat_dseg', 'inputnode.label_file')]),
            (bold_fit_wf, t2s_reporting_wf, [
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
                ('outputnode.coreg_boldref', 'inputnode.boldref'),
            ]),
            (bold_native_wf, t2s_reporting_wf, [
                ('outputnode.t2star_map', 'inputnode.t2star_file'),
            ]),
            (t2s_reporting_wf, ds_report_t2scomp, [('outputnode.t2s_comp_report', 'in_file')]),
            (t2s_reporting_wf, ds_report_t2star_hist, [('outputnode.t2star_hist', 'in_file')]),
        ])  # fmt:skip

    if config.workflow.level == 'resampling':
        # Fill-in datasinks of reportlets seen so far
        for node in workflow.list_node_names():
            if node.split('.')[-1].startswith('ds_report'):
                workflow.get_node(node).inputs.base_directory = output_dir
                workflow.get_node(node).inputs.source_file = bold_file
        return workflow

    # Resample to anatomical space
    bold_anat_wf = init_bold_volumetric_resample_wf(
        metadata=all_metadata[0],
        fieldmap_id=fieldmap_id if not multiecho else None,
        omp_nthreads=omp_nthreads,
        mem_gb=mem_gb,
        jacobian='fmap-jacobian' not in config.workflow.ignore,
        name='bold_anat_wf',
    )
    bold_anat_wf.inputs.inputnode.resolution = 'native'

    workflow.connect([
        (inputnode, bold_anat_wf, [
            ('anat_preproc', 'inputnode.target_ref_file'),
            ('anat_mask', 'inputnode.target_mask'),
            ('fmap_ref', 'inputnode.fmap_ref'),
            ('fmap_coeff', 'inputnode.fmap_coeff'),
            ('fmap_id', 'inputnode.fmap_id'),
        ]),
        (bold_fit_wf, bold_anat_wf, [
            ('outputnode.coreg_boldref', 'inputnode.bold_ref_file'),
            ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
            ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
        ]),
        (bold_native_wf, bold_anat_wf, [
            ('outputnode.bold_minimal', 'inputnode.bold_file'),
            ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
        ]),
    ])  # fmt:skip

    # Full derivatives, including resampled BOLD series
    if nonstd_spaces.intersection(('anat', 'T1w', 'T2w')):
        ds_bold_anat_wf = init_ds_volumes_wf(
            bids_root=str(config.execution.bids_dir),
            output_dir=output_dir,
            multiecho=multiecho,
            metadata=all_metadata[0],
            name='ds_bold_anat_wf',
        )
        ds_bold_anat_wf.inputs.inputnode.source_files = bold_series
        ds_bold_anat_wf.inputs.inputnode.space = reference_anat

        workflow.connect([
            (bold_fit_wf, ds_bold_anat_wf, [
                ('outputnode.bold_mask', 'inputnode.bold_mask'),
                ('outputnode.coreg_boldref', 'inputnode.bold_ref'),
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
                ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
                ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
            ]),
            (bold_native_wf, ds_bold_anat_wf, [('outputnode.t2star_map', 'inputnode.t2star')]),
            (bold_anat_wf, ds_bold_anat_wf, [
                ('outputnode.bold_file', 'inputnode.bold'),
                ('outputnode.resampling_reference', 'inputnode.ref_file'),
            ]),
        ])  # fmt:skip

    if spaces.cached.get_spaces(nonstandard=False, dim=(3,)):
        # Missing:
        #  * Clipping BOLD after resampling
        #  * Resampling parcellations
        bold_std_wf = init_bold_volumetric_resample_wf(
            metadata=all_metadata[0],
            fieldmap_id=fieldmap_id if not multiecho else None,
            omp_nthreads=omp_nthreads,
            mem_gb=mem_gb,
            jacobian='fmap-jacobian' not in config.workflow.ignore,
            name='bold_std_wf',
        )
        ds_bold_std_wf = init_ds_volumes_wf(
            bids_root=str(config.execution.bids_dir),
            output_dir=output_dir,
            multiecho=multiecho,
            metadata=all_metadata[0],
            name='ds_bold_std_wf',
        )
        ds_bold_std_wf.inputs.inputnode.source_files = bold_series

        workflow.connect([
            (inputnode, bold_std_wf, [
                ('std_t1w', 'inputnode.target_ref_file'),
                ('std_mask', 'inputnode.target_mask'),
                ('anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('std_resolution', 'inputnode.resolution'),
                ('fmap_ref', 'inputnode.fmap_ref'),
                ('fmap_coeff', 'inputnode.fmap_coeff'),
                ('fmap_id', 'inputnode.fmap_id'),
            ]),
            (bold_fit_wf, bold_std_wf, [
                ('outputnode.coreg_boldref', 'inputnode.bold_ref_file'),
                ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
            ]),
            (bold_native_wf, bold_std_wf, [
                ('outputnode.bold_minimal', 'inputnode.bold_file'),
                ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
            ]),
            (inputnode, ds_bold_std_wf, [
                ('anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('std_t1w', 'inputnode.template'),
                ('std_space', 'inputnode.space'),
                ('std_resolution', 'inputnode.resolution'),
                ('std_cohort', 'inputnode.cohort'),
            ]),
            (bold_fit_wf, ds_bold_std_wf, [
                ('outputnode.bold_mask', 'inputnode.bold_mask'),
                ('outputnode.coreg_boldref', 'inputnode.bold_ref'),
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
                ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
                ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
            ]),
            (bold_native_wf, ds_bold_std_wf, [('outputnode.t2star_map', 'inputnode.t2star')]),
            (bold_std_wf, ds_bold_std_wf, [
                ('outputnode.bold_file', 'inputnode.bold'),
                ('outputnode.resampling_reference', 'inputnode.ref_file'),
            ]),
        ])  # fmt:skip

    if config.workflow.surface_recon_method and freesurfer_spaces:
        workflow.__postdesc__ += """\
Non-gridded (surface) resamplings were performed using `mri_vol2surf`
(FreeSurfer).
"""
        config.loggers.workflow.debug('Creating BOLD surface-sampling workflow.')
        bold_surf_wf = init_bold_surf_wf(
            mem_gb=mem_gb['resampled'],
            surface_spaces=freesurfer_spaces,
            medial_surface_nan=config.workflow.medial_surface_nan,
            metadata=all_metadata[0],
            output_dir=output_dir,
            name='bold_surf_wf',
        )
        bold_surf_wf.inputs.inputnode.source_file = bold_file
        workflow.connect([
            (inputnode, bold_surf_wf, [
                ('subjects_dir', 'inputnode.subjects_dir'),
                ('subject_id', 'inputnode.subject_id'),
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (bold_anat_wf, bold_surf_wf, [('outputnode.bold_file', 'inputnode.bold_anat')]),
        ])  # fmt:skip

        # sources are bold_file, motion_xfm, boldref2anat_xfm, fsnative2anat_xfm
        merge_surface_sources = pe.Node(
            niu.Merge(4),
            name='merge_surface_sources',
            run_without_submitting=True,
        )
        merge_surface_sources.inputs.in1 = bold_file
        workflow.connect([
            (bold_fit_wf, merge_surface_sources, [
                ('outputnode.motion_xfm', 'in2'),
                ('outputnode.boldref2anat_xfm', 'in3'),
            ]),
            (inputnode, merge_surface_sources, [
                ('fsnative2anat_xfm', 'in4'),
            ]),
            (merge_surface_sources, bold_surf_wf, [
                ('out', 'inputnode.sources'),
            ]),
        ])  # fmt:skip

    cifti_output = config.workflow.cifti_output
    if cifti_output:
        from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms

        from nibabies.workflows.bold.alignment import (
            init_subcortical_mni_alignment_wf,
            init_subcortical_rois_wf,
        )
        from nibabies.workflows.bold.resampling import (
            init_bold_fsLR_resampling_wf,
            init_bold_grayords_wf,
            init_goodvoxels_bold_mask_wf,
        )

        bold_MNIInfant_wf = init_bold_volumetric_resample_wf(
            metadata=all_metadata[0],
            fieldmap_id=fieldmap_id if not multiecho else None,
            omp_nthreads=omp_nthreads,
            mem_gb=mem_gb,
            jacobian='fmap-jacobian' not in config.workflow.ignore,
            name='bold_MNIInfant_wf',
        )

        bold_fsLR_resampling_wf = init_bold_fsLR_resampling_wf(
            grayord_density=cifti_output,
            omp_nthreads=omp_nthreads,
            mem_gb=mem_gb['resampled'],
        )

        if config.workflow.project_goodvoxels:
            goodvoxels_bold_mask_wf = init_goodvoxels_bold_mask_wf(mem_gb['resampled'])

            workflow.connect([
                (inputnode, goodvoxels_bold_mask_wf, [('anat_ribbon', 'inputnode.anat_ribbon')]),
                (bold_anat_wf, goodvoxels_bold_mask_wf, [
                    ('outputnode.bold_file', 'inputnode.bold_file'),
                ]),
                (goodvoxels_bold_mask_wf, bold_fsLR_resampling_wf, [
                    ('outputnode.goodvoxels_mask', 'inputnode.volume_roi'),
                ]),
            ])  # fmt:skip

            bold_fsLR_resampling_wf.__desc__ += """\
A "goodvoxels" mask was applied during volume-to-surface sampling in fsLR space,
excluding voxels whose time-series have a locally high coefficient of variation.
"""

        # MNIInfant -> MNI6 registrations (per ROI)
        MNIInfant_aseg = pe.Node(
            ApplyTransforms(interpolation='MultiLabel'),
            name='MNIInfant_aseg',
            mem_gb=1,
        )

        subcortical_rois_wf = init_subcortical_rois_wf()
        subcortical_mni_alignment_wf = init_subcortical_mni_alignment_wf()

        bold_grayords_wf = init_bold_grayords_wf(
            grayord_density=cifti_output,
            repetition_time=all_metadata[0]['RepetitionTime'],
        )

        ds_bold_cifti = pe.Node(
            DerivativesDataSink(
                base_directory=output_dir,
                dismiss_entities=DEFAULT_DISMISS_ENTITIES,
                space='fsLR',
                density=cifti_output,
                suffix='bold',
                compress=False,
                TaskName=all_metadata[0].get('TaskName'),
                **prepare_timing_parameters(all_metadata[0]),
            ),
            name='ds_bold_cifti',
            run_without_submitting=True,
        )
        ds_bold_cifti.inputs.source_file = bold_file

        mniinfant_res = 2 if cifti_output == '91k' else 1
        inputnode.inputs.mniinfant_mask = get_MNIInfant_mask(spaces, mniinfant_res)

        workflow.connect([
            # Resample BOLD to MNI152NLin6Asym, may duplicate bold_std_wf above
            (inputnode, bold_MNIInfant_wf, [
                ('mniinfant_mask', 'inputnode.target_ref_file'),
                ('mniinfant_mask', 'inputnode.target_mask'),
                ('anat2mniinfant_xfm', 'inputnode.anat2std_xfm'),
                ('fmap_ref', 'inputnode.fmap_ref'),
                ('fmap_coeff', 'inputnode.fmap_coeff'),
                ('fmap_id', 'inputnode.fmap_id'),
            ]),
            (bold_fit_wf, bold_MNIInfant_wf, [
                ('outputnode.coreg_boldref', 'inputnode.bold_ref_file'),
                ('outputnode.boldref2fmap_xfm', 'inputnode.boldref2fmap_xfm'),
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
            ]),
            (bold_native_wf, bold_MNIInfant_wf, [
                ('outputnode.bold_minimal', 'inputnode.bold_file'),
                ('outputnode.motion_xfm', 'inputnode.motion_xfm'),
            ]),
            (inputnode, MNIInfant_aseg, [
                ('anat2mniinfant_xfm', 'transforms'),
                ('anat_aseg', 'input_image'),
            ]),
            (bold_MNIInfant_wf, MNIInfant_aseg, [
                ('outputnode.resampling_reference', 'reference_image'),
            ]),
            (bold_MNIInfant_wf, subcortical_mni_alignment_wf, [
                ('outputnode.bold_file', 'inputnode.MNIInfant_bold'),
            ]),
            (MNIInfant_aseg, subcortical_rois_wf, [
                ('output_image', 'inputnode.MNIInfant_aseg'),
            ]),
            (subcortical_rois_wf, subcortical_mni_alignment_wf, [
                ('outputnode.MNIInfant_rois', 'inputnode.MNIInfant_rois'),
                ('outputnode.MNI152_rois', 'inputnode.MNI152_rois'),
            ]),
            (subcortical_mni_alignment_wf, bold_grayords_wf, [
                ('outputnode.subcortical_volume', 'inputnode.bold_std'),
                ('outputnode.subcortical_labels', 'inputnode.bold_labels'),
            ]),

            # Resample anat-space BOLD to fsLR surfaces
            (inputnode, bold_fsLR_resampling_wf, [
                ('white', 'inputnode.white'),
                ('pial', 'inputnode.pial'),
                ('midthickness', 'inputnode.midthickness'),
                ('midthickness_fsLR', 'inputnode.midthickness_fsLR'),
                ('sphere_reg_fsLR', 'inputnode.sphere_reg_fsLR'),
                ('cortex_mask', 'inputnode.cortex_mask'),
            ]),
            (bold_anat_wf, bold_fsLR_resampling_wf, [
                ('outputnode.bold_file', 'inputnode.bold_file'),
            ]),
            # (bold_MNI6_wf, bold_grayords_wf, [
            #     ('outputnode.bold_file', 'inputnode.bold_std'),
            # ]),
            (bold_fsLR_resampling_wf, bold_grayords_wf, [
                ('outputnode.bold_fsLR', 'inputnode.bold_fsLR'),
            ]),
            (bold_grayords_wf, ds_bold_cifti, [
                ('outputnode.dtseries', 'in_file'),
                (('outputnode.dtseries_metadata', _read_json), 'meta_dict'),
            ]),
        ])  # fmt:skip

    bold_confounds_wf = init_bold_confs_wf(
        mem_gb=mem_gb['largemem'],
        metadata=all_metadata[0],
        freesurfer=config.workflow.run_reconall,
        regressors_all_comps=config.workflow.regressors_all_comps,
        regressors_fd_th=config.workflow.regressors_fd_th,
        regressors_dvars_th=config.workflow.regressors_dvars_th,
        name='bold_confounds_wf',
    )

    ds_confounds = pe.Node(
        DerivativesDataSink(
            base_directory=output_dir,
            desc='confounds',
            suffix='timeseries',
            dismiss_entities=DEFAULT_DISMISS_ENTITIES,
        ),
        name='ds_confounds',
        run_without_submitting=True,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )
    ds_confounds.inputs.source_file = bold_file

    workflow.connect([
        (inputnode, bold_confounds_wf, [
            ('anat_tpms', 'inputnode.anat_tpms'),
            ('anat_mask', 'inputnode.anat_mask'),
        ]),
        (bold_fit_wf, bold_confounds_wf, [
            ('outputnode.bold_mask', 'inputnode.bold_mask'),
            ('outputnode.movpar_file', 'inputnode.movpar_file'),
            ('outputnode.rmsd_file', 'inputnode.rmsd_file'),
            ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
            ('outputnode.dummy_scans', 'inputnode.skip_vols'),
        ]),
        (bold_native_wf, bold_confounds_wf, [
            ('outputnode.bold_native', 'inputnode.bold'),
        ]),
        (bold_confounds_wf, ds_confounds, [
            ('outputnode.confounds_file', 'in_file'),
            ('outputnode.confounds_metadata', 'meta_dict'),
        ]),
    ])  # fmt:skip

    # MG: Carpetplot workflow only work with CIFTI
    if cifti_output:
        carpetplot_wf = init_carpetplot_wf(
            mem_gb=mem_gb['resampled'],
            metadata=all_metadata[0],
            cifti_output=cifti_output,
            name='carpetplot_wf',
        )

        def _last(inlist):
            return inlist[-1]

        workflow.connect([
            (bold_grayords_wf, carpetplot_wf, [
                ('outputnode.dtseries', 'inputnode.cifti_bold'),
            ]),
            (bold_fit_wf, carpetplot_wf, [
                ('outputnode.dummy_scans', 'inputnode.dummy_scans'),
                ('outputnode.bold_mask', 'inputnode.bold_mask'),
                ('outputnode.boldref2anat_xfm', 'inputnode.boldref2anat_xfm'),
            ]),
            (bold_native_wf, carpetplot_wf, [
                ('outputnode.bold_native', 'inputnode.bold'),
            ]),
            (bold_confounds_wf, carpetplot_wf, [
                ('outputnode.confounds_file', 'inputnode.confounds_file'),
                ('outputnode.crown_mask', 'inputnode.crown_mask'),
                (('outputnode.acompcor_masks', _last), 'inputnode.acompcor_mask'),
            ]),
        ])  # fmt:skip

    # Fill-in datasinks of reportlets seen so far
    for node in workflow.list_node_names():
        if node.split('.')[-1].startswith('ds_report'):
            workflow.get_node(node).inputs.base_directory = output_dir
            workflow.get_node(node).inputs.source_file = bold_file

    return workflow


def _get_wf_name(bold_fname, prefix):
    """
    Derive the workflow name for supplied BOLD file.

    >>> _get_wf_name("/completely/made/up/path/sub-01_task-nback_bold.nii.gz", "bold")
    'bold_task_nback_wf'
    >>> _get_wf_name(
    ...     "/completely/made/up/path/sub-01_task-nback_run-01_echo-1_bold.nii.gz",
    ...     "preproc",
    ... )
    'preproc_task_nback_run_01_echo_1_wf'

    """
    from nipype.utils.filemanip import split_filename

    fname = split_filename(bold_fname)[1]
    fname_nosub = '_'.join(fname.split('_')[1:-1])
    return f'{prefix}_{fname_nosub.replace("-", "_")}_wf'


def extract_entities(file_list):
    """
    Return a dictionary of common entities given a list of files.

    Examples
    --------
    >>> extract_entities("sub-01/anat/sub-01_T1w.nii.gz")
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(["sub-01/anat/sub-01_T1w.nii.gz"] * 2)
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(["sub-01/anat/sub-01_run-1_T1w.nii.gz",
    ...                   "sub-01/anat/sub-01_run-2_T1w.nii.gz"])
    {'subject': '01', 'run': [1, 2], 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}

    """
    from collections import defaultdict

    from bids.layout import parse_file_entities

    entities = defaultdict(list)
    for e, v in [
        ev_pair for f in listify(file_list) for ev_pair in parse_file_entities(f).items()
    ]:
        entities[e].append(v)

    def _unique(inlist):
        inlist = sorted(set(inlist))
        if len(inlist) == 1:
            return inlist[0]
        return inlist

    return {k: _unique(v) for k, v in entities.items()}


def _read_json(in_file):
    from json import loads
    from pathlib import Path

    return loads(Path(in_file).read_text())


def get_MNIInfant_mask(spaces: 'SpatialReferences', res: str | int) -> str:
    """Parse spaces and return matching MNIInfant space, including cohort."""
    import templateflow.api as tf

    for ref in spaces.references:
        if ref.space == 'MNIInfant' and f'res-{res}' in str(ref):
            return str(
                tf.get(
                    'MNIInfant',
                    cohort=ref.spec['cohort'],
                    resolution=res,
                    desc='brain',
                    suffix='mask',
                )
            )

    raise FileNotFoundError(f'MNIInfant mask (resolution {res}) not found.')



"""BIDS-related interfaces."""

from pathlib import Path

from bids.utils import listify
from nipype.interfaces.base import (
    DynamicTraitedSpec,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.interfaces.io import add_traits
from nipype.interfaces.utility.base import _ravel


class _BIDSURIInputSpec(DynamicTraitedSpec):
    dataset_links = traits.Dict(mandatory=True, desc='Dataset links')
    out_dir = traits.Str(mandatory=True, desc='Output directory')


class _BIDSURIOutputSpec(TraitedSpec):
    out = traits.List(
        traits.Str,
        desc='BIDS URI(s) for file',
    )


class BIDSURI(SimpleInterface):
    """Convert input filenames to BIDS URIs, based on links in the dataset.

    This interface can combine multiple lists of inputs.
    """

    input_spec = _BIDSURIInputSpec
    output_spec = _BIDSURIOutputSpec

    def __init__(self, numinputs=0, **inputs):
        super().__init__(**inputs)
        self._numinputs = numinputs
        if numinputs >= 1:
            input_names = [f'in{i + 1}' for i in range(numinputs)]
        else:
            input_names = []
        add_traits(self.inputs, input_names)

    def _run_interface(self, runtime):
        inputs = [getattr(self.inputs, f'in{i + 1}') for i in range(self._numinputs)]
        in_files = listify(inputs)
        in_files = _ravel(in_files)
        # Remove undefined inputs
        in_files = [f for f in in_files if isdefined(f)]
        # Convert the dataset links to BIDS URI prefixes
        updated_keys = {f'bids:{k}:': Path(v) for k, v in self.inputs.dataset_links.items()}
        updated_keys['bids::'] = Path(self.inputs.out_dir)
        # Convert the paths to BIDS URIs
        out = [_find_nearest_path(updated_keys, f) for f in in_files]
        self._results['out'] = out

        return runtime


def _find_nearest_path(paths: dict[str, Path], input_path: str | Path):
    """Find the nearest relative path from an input path to a dictionary of paths.

    If ``input_path`` is not relative to any of the paths in ``path_dict``,
    the absolute path string is returned.

    If ``input_path`` is already a BIDS-URI, then it will be returned unmodified.

    Parameters
    ----------
    path_dict : dict of (str, Path)
        A dictionary of paths.
    input_path : Path
        The input path to match.

    Returns
    -------
    matching_path : str
        The nearest relative path from the input path to a path in the dictionary.
        This is either the concatenation of the associated key from ``path_dict``
        and the relative path from the associated value from ``path_dict`` to ``input_path``,
        or the absolute path to ``input_path`` if no matching path is found from ``path_dict``.

    Examples
    --------
    >>> from pathlib import Path
    >>> path_dict = {
    ...     'bids::': Path('/data/derivatives/fmriprep'),
    ...     'bids:raw:': Path('/data'),
    ...     'bids:deriv-0:': Path('/data/derivatives/source-1'),
    ... }
    >>> input_path = Path('/data/derivatives/source-1/sub-01/func/sub-01_task-rest_bold.nii.gz')
    >>> _find_nearest_path(path_dict, input_path)  # match to 'bids:deriv-0:'
    'bids:deriv-0:sub-01/func/sub-01_task-rest_bold.nii.gz'
    >>> input_path = Path('/out/sub-01/func/sub-01_task-rest_bold.nii.gz')
    >>> _find_nearest_path(path_dict, input_path)  # no match- absolute path
    '/out/sub-01/func/sub-01_task-rest_bold.nii.gz'
    >>> input_path = Path('/data/sub-01/func/sub-01_task-rest_bold.nii.gz')
    >>> _find_nearest_path(path_dict, input_path)  # match to 'bids:raw:'
    'bids:raw:sub-01/func/sub-01_task-rest_bold.nii.gz'
    >>> input_path = 'bids::sub-01/func/sub-01_task-rest_bold.nii.gz'
    >>> _find_nearest_path(path_dict, input_path)  # already a BIDS-URI
    'bids::sub-01/func/sub-01_task-rest_bold.nii.gz'
    """
    # Don't modify BIDS-URIs
    if isinstance(input_path, str) and input_path.startswith('bids:'):
        return input_path

    input_path = Path(input_path)
    matching_path = None
    for key, path in paths.items():
        if input_path.is_relative_to(path):
            relative_path = input_path.relative_to(path)
            if (matching_path is None) or (len(relative_path.parts) < len(matching_path.parts)):
                matching_key = key
                matching_path = relative_path

    if matching_path is None:
        matching_path = str(input_path.absolute())
    else:
        matching_path = f'{matching_key}{matching_path}'

    return matching_path



"""
Subcortical alignment into MNI space
"""

from nipype.interfaces import fsl
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.engine.workflows import LiterateWorkflow as Workflow

from nibabies.data import load as load_data
from nibabies.interfaces.workbench import VolumeLabelImport


def init_subcortical_rois_wf(*, name: str = 'subcortical_rois_wf'):
    """
    Refine segmentations into volumes of expected CIFTI subcortical structures.


    Parameters
    ----------
    name : :obj:`str`
        Name of the workflow

    Inputs
    ------
    MNIInfant_aseg : :obj:`str`
        FreeSurfer's aseg in MNIInfant space

    Outputs
    -------
    MNIInfant_rois : :obj:`str`
        Subcortical ROIs in MNIInfant space
    MNI152_rois : :obj:`str`
        Subcortical ROIs in `MNI152NLin6Asym` space
    """
    from niworkflows.interfaces.nibabel import MapLabels
    from templateflow.api import get as get_template

    # TODO: Implement BOLD refinement once InfantFS outputs subj/mri/wmparc.mgz
    # The code is found at
    # https://github.com/DCAN-Labs/dcan-infant-pipeline/blob/
    # 0e9c2fe32fb4a5032d0a2a3e0905ad97fa52b398/PostFreeSurfer/scripts/
    # FreeSurfer2CaretConvertAndRegisterNonlinear.sh
    # Lines 70-78 & 116-127
    # #
    # For now, just use the aseg

    workflow = Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=['MNIInfant_aseg']), name='inputnode')
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['MNIInfant_rois', 'MNI152_rois']),
        name='outputnode',
    )
    # Fetch the HCP volumetric template
    tpl_rois = get_template(
        'MNI152NLin6Asym', resolution=2, atlas='HCP', suffix='dseg', raise_empty=True
    )
    outputnode.inputs.MNI152_rois = tpl_rois

    # This will only used for the wmparc in subject space
    # For now, define it and don't run it
    # TODO: Move to TemplateFlow

    # tpl_avgwmparc = load_data("tpl-MNI152NLin6Asym_res-01_desc-avgwmparc_dseg.nii.gz")
    # applywarp_tpl = pe.Node(
    #     fsl.ApplyWarp(in_file=tpl_avgwmparc, ref_file=tpl_rois, interp="nn"),
    #     name="applywarp_std"
    # )

    map_labels = pe.Node(
        MapLabels(mappings_file=load_data('FreeSurferLabelRemappings.json')),
        name='map_labels',
    )

    refine_bold_rois = pe.Node(
        VolumeLabelImport(
            label_list_file=load_data('FreeSurferSubcorticalLabelTableLut.txt'),
            discard_others=True,
        ),
        name='refine_bold_rois',
    )

    # fmt: off
    workflow.connect([
        (inputnode, map_labels, [('MNIInfant_aseg', 'in_file')]),
        (map_labels, refine_bold_rois, [('out_file', 'in_file')]),
        # (applywarp_tpl, refine_std_rois, [("out_file", "in_file")]),
        (refine_bold_rois, outputnode, [('out_file', 'MNIInfant_rois')]),
    ])
    # fmt: on
    return workflow


def init_subcortical_mni_alignment_wf(
    *,
    vol_sigma: float = 0.8,
    name: str = 'subcortical_mni_alignment_wf',
):
    """
    Align individual subcortical structures into MNI space.

    This is a nipype workflow port of the DCAN infant pipeline:
    https://github.com/DCAN-Labs/dcan-infant-pipeline/blob\
    /master/fMRISurface/scripts/SubcorticalAlign_ROIs.sh


    Parameters
    ----------
    name : :obj:`str`
        Name of the workflow
    vol_sigma : :obj:`float`
        The sigma for the gaussian volume smoothing kernel, in mm

    Inputs
    ------
    MNIInfant_bold : :obj:`str`
        BOLD file in MNI Infant space
    MNIInfant_rois : :obj:`str`
        File containing ROIs in MNI Infant space
    MNI152_rois : :obj:`str`
        File containing ROIs in MNI152NLin6Asym space

    Outputs
    -------
    subcortical_volume : :obj:`str`
        Volume file containing all ROIs individually aligned to standard
    subcortical_labels : :obj:`str`
        Volume file containing all labels
    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.nibabel import MergeROIs

    from ...interfaces.workbench import (
        CiftiCreateDenseTimeseries,
        CiftiCreateLabel,
        CiftiDilate,
        CiftiResample,
        CiftiSeparate,
        CiftiSmooth,
        VolumeAffineResample,
        VolumeAllLabelsToROIs,
        VolumeLabelExportTable,
    )

    # reuse saved atlas to atlas transform
    atlas_xfm = load_data('MNIInfant_to_MNI1526NLinAsym.mat')
    inputnode = pe.Node(
        niu.IdentityInterface(fields=['MNIInfant_bold', 'MNIInfant_rois', 'MNI152_rois']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['subcortical_volume', 'subcortical_labels']),
        name='outputnode',
    )

    applyxfm_atlas = pe.Node(fsl.ApplyXFM(in_matrix_file=atlas_xfm), name='applyxfm_atlas')
    vol_resample = pe.Node(
        VolumeAffineResample(method='ENCLOSING_VOXEL', flirt=True, affine=atlas_xfm),
        name='vol_resample',
    )
    subj_rois = pe.Node(VolumeAllLabelsToROIs(label_map=1), name='subj_rois')
    split_rois = pe.Node(fsl.Split(dimension='t'), name='split_rois')
    atlas_rois = pe.Node(VolumeAllLabelsToROIs(label_map=1), name='atlas_rois')
    split_atlas_rois = pe.Node(fsl.Split(dimension='t'), name='split_atlas_rois')
    atlas_labels = pe.Node(VolumeLabelExportTable(label_map=1), name='atlas_labels')
    parse_labels = pe.Node(
        niu.Function(function=parse_roi_labels, output_names=['structures', 'label_ids']),
        name='parse_labels',
    )

    # The following is wrapped in a for-loop, iterating across each roi
    # Instead, we will use MapNodes and iter across the varying inputs
    roi2atlas = pe.MapNode(
        fsl.FLIRT(
            searchr_x=[-20, 20],
            searchr_y=[-20, 20],
            searchr_z=[-20, 20],
            interp='nearestneighbour',
        ),
        name='roi2atlas',
        iterfield=['in_file', 'reference'],
    )
    applyxfm_roi = pe.MapNode(
        fsl.ApplyXFM(interp='spline'),
        iterfield=['reference', 'in_matrix_file'],
        name='applyxfm_roi',
        mem_gb=4,
    )
    bold_mask_roi = pe.MapNode(
        fsl.ApplyMask(),
        iterfield=['in_file', 'mask_file'],
        name='bold_mask_roi',
    )
    mul_roi = pe.MapNode(
        fsl.BinaryMaths(operation='mul'),
        iterfield=['in_file', 'operand_value'],
        name='mul_roi',
    )
    mul_atlas_roi = pe.MapNode(
        fsl.BinaryMaths(operation='mul'),
        iterfield=['in_file', 'operand_value'],
        name='mul_atlas_roi',
    )
    vol_label = pe.MapNode(
        VolumeLabelImport(drop_unused_labels=True),
        iterfield=['in_file'],
        name='vol_label',
    )
    vol_atlas_label = pe.MapNode(
        VolumeLabelImport(drop_unused_labels=True),
        iterfield=['in_file'],
        name='vol_atlas_label',
    )
    create_dtseries = pe.MapNode(
        CiftiCreateDenseTimeseries(),
        iterfield=['volume_data', 'volume_structure_labels'],
        name='create_dtseries',
    )
    create_label = pe.MapNode(
        CiftiCreateLabel(),
        iterfield=['volume_label', 'structure_label_volume'],
        name='create_label',
    )
    dilate = pe.MapNode(
        CiftiDilate(direction='COLUMN', surface_distance=0, volume_distance=10),
        iterfield=['in_file'],
        name='dilate',
    )
    resample = pe.MapNode(
        CiftiResample(
            direction='COLUMN',
            template_direction='COLUMN',
            surface_method='ADAP_BARY_AREA',
            volume_method='CUBIC',
            volume_predilate=10,
        ),
        iterfield=['in_file', 'template'],
        name='resample',
    )
    smooth = pe.MapNode(
        CiftiSmooth(direction='COLUMN', fix_zeros_vol=True, sigma_surf=0, sigma_vol=vol_sigma),
        iterfield=['in_file'],
        name='smooth',
    )
    separate = pe.MapNode(
        CiftiSeparate(direction='COLUMN', volume_all_file='volume_all.nii.gz'),
        iterfield=['in_file'],
        name='separate',
    )
    fmt_agg_rois = pe.Node(
        niu.Function(
            function=format_agg_rois,
            output_names=['first_image', 'op_files', 'op_string'],
        ),
        name='fmt_agg_rois',
    )
    agg_rois = pe.Node(fsl.MultiImageMaths(), name='agg_rois')
    merge_rois = pe.Node(MergeROIs(), name='merge_rois')

    workflow = Workflow(name=name)
    workflow.connect([
        (inputnode, applyxfm_atlas, [
            ('MNIInfant_bold', 'in_file'),
            ('MNI152_rois', 'reference')]),
        (inputnode, vol_resample, [
            ('MNIInfant_rois', 'in_file'),
            ('MNIInfant_rois', 'flirt_source_volume')]),
        (applyxfm_atlas, vol_resample, [
            ('out_file', 'volume_space'),
            ('out_file', 'flirt_target_volume')]),
        (inputnode, subj_rois, [('MNIInfant_rois', 'in_file')]),
        (inputnode, atlas_rois, [('MNI152_rois', 'in_file')]),
        (subj_rois, split_rois, [('out_file', 'in_file')]),
        (atlas_rois, split_atlas_rois, [('out_file', 'in_file')]),
        (inputnode, atlas_labels, [('MNI152_rois', 'in_file')]),
        (atlas_labels, parse_labels, [('out_file', 'label_file')]),
        # for loop across ROIs
        (split_rois, roi2atlas, [('out_files', 'in_file')]),
        (split_atlas_rois, roi2atlas, [('out_files', 'reference')]),
        (inputnode, applyxfm_roi, [('MNIInfant_bold', 'in_file')]),
        (split_atlas_rois, applyxfm_roi, [('out_files', 'reference')]),
        (roi2atlas, applyxfm_roi, [('out_matrix_file', 'in_matrix_file')]),
        (applyxfm_roi, bold_mask_roi, [('out_file', 'in_file')]),
        (roi2atlas, bold_mask_roi, [('out_file', 'mask_file')]),
        (roi2atlas, mul_roi, [('out_file', 'in_file')]),
        (parse_labels, mul_roi, [('label_ids', 'operand_value')]),
        (split_atlas_rois, mul_atlas_roi, [('out_files', 'in_file')]),
        (parse_labels, mul_atlas_roi, [('label_ids', 'operand_value')]),
        (mul_roi, vol_label, [('out_file', 'in_file')]),
        (atlas_labels, vol_label, [('out_file', 'label_list_file')]),
        (mul_atlas_roi, vol_atlas_label, [('out_file', 'in_file')]),
        (atlas_labels, vol_atlas_label, [('out_file', 'label_list_file')]),
        (bold_mask_roi, create_dtseries, [('out_file', 'volume_data')]),
        (vol_label, create_dtseries, [('out_file', 'volume_structure_labels')]),
        (vol_atlas_label, create_label, [
            ('out_file', 'volume_label'),
            ('out_file', 'structure_label_volume')]),
        (create_dtseries, dilate, [('out_file', 'in_file')]),
        (dilate, resample, [('out_file', 'in_file')]),
        (create_label, resample, [('out_file', 'template')]),
        (resample, smooth, [('out_file', 'in_file')]),
        (smooth, separate, [('out_file', 'in_file')]),
        # end loop
        (mul_roi, fmt_agg_rois, [('out_file', 'rois')]),
        (fmt_agg_rois, agg_rois, [
            ('first_image', 'in_file'),
            ('op_files', 'operand_files'),
            ('op_string', 'op_string')]),
        (separate, merge_rois, [('volume_all_file', 'in_files')]),
        (merge_rois, outputnode, [('out_file', 'subcortical_volume')]),
        (inputnode, outputnode, [('MNI152_rois', 'subcortical_labels')]),
    ])  # fmt:skip
    return workflow


def parse_roi_labels(label_file: str):
    """
    Parse a label file composed of one or more sets of:
    <labelname>
    <key> <red> <green> <blue> <alpha>

    Return a list of structure names and label keys.

    >>> structs, ids = parse_roi_labels(datadir / "FreeSurferSubcorticalLabelTableLut.txt")
    >>> structs
    ['ACCUMBENS_LEFT', 'ACCUMBENS_RIGHT', 'AMYGDALA_LEFT', ...]
    >>> ids
    [26, 58, 18, ...]
    """

    with open(label_file) as fp:
        lines = fp.readlines()
    if len(lines) % 2 == 1:
        raise RuntimeError('Label file is incomplete or invalid')
    structs, label_ids = [], []
    for idx, line in enumerate(lines):
        if idx % 2 == 0:
            structs.append(line.strip())
        else:
            label_ids.append(int(line.split(' ', 1)[0]))
    return structs, label_ids


def format_agg_rois(rois):
    """
    Helper function to format MultiImageMaths command.

    Parameters
    ----------
    rois : `list` of `str`s
        List of files

    Returns
    -------
    first_image
    op_files
    op_string

    """
    return rois[0], rois[1:], ('-add %s ' * (len(rois) - 1)).strip()



from pathlib import Path

import pytest

from nibabies.utils.bids import _get_age_from_tsv


def create_tsv(data: dict, out_file: Path) -> None:
    import pandas as pd

    pd.DataFrame(data).to_csv(out_file, index=False, sep='\t')


def create_sidecar(tsv_file: Path, units) -> None:
    import json

    out_file = tsv_file.with_suffix('.json')
    data = {'age': {'Units': units}}
    out_file.write_text(json.dumps(data))


age = {'age': [4, 4, 4]}
age_weeks = {'age_weeks': [4, 8, 12]}
age_months = {'age_months': [3, 6, 9]}
age_years = {'age_years': [1, 1, 2]}

participants = {'participant_id': ['sub-1', 'sub-2', 'sub-11']}
sessions = {'session_id': ['ses-1', 'ses-2', 'ses-3']}
scans = {
    'filename': [
        'dwi/sub-01_dwi.nii.gz',
        'anat/sub-01_T1w.nii.gz',
        'func/sub-01_task-rest_bold.nii.gz',
    ]
}


@pytest.mark.parametrize(
    ('idx_col', 'idx_val', 'data', 'units', 'expected'),
    [
        ('session_id', 'ses-1', age, 'months', 4),
        ('session_id', 'ses-1', age, 'weeks', 1),  # Convert from 4 weeks -> 1 month
        ('session_id', 'ses-2', age_weeks, False, 2),
        ('participant_id', 'sub-1', age_months, False, 3),
        ('participant_id', 'sub-11', age_years, False, 24),
        ('session_id', 'ses-3', {**age_months, **age}, False, 9),
        ('filename', r'^anat.*', age_months, False, 6),
    ],
)
def test_get_age_from_tsv(tmp_path, idx_col, idx_val, data, units, expected):
    tsv_file = tmp_path / 'test-age-parsing.tsv'

    if idx_col == 'participant_id':
        base = participants
    elif idx_col == 'session_id':
        base = sessions
    elif idx_col == 'filename':
        base = scans

    create_tsv({**base, **data}, tsv_file)
    if units:
        create_sidecar(tsv_file, units)

    res = _get_age_from_tsv(tsv_file, idx_col, idx_val)
    assert res == expected


def test_get_age_from_tsv_error(tmp_path):
    tsv_file = tmp_path / 'participants.tsv'

    create_tsv({**participants, **age}, tsv_file)
    with pytest.raises(FileNotFoundError):
        _get_age_from_tsv(tsv_file, 'participant_id', 'sub-1')


def test_get_age_from_tsv_warning(tmp_path):
    tsv_file = tmp_path / 'participants.tsv'
    dual_participants = {'participant_id': ['sub-1', 'sub-2', 'sub-2']}
    create_tsv({**dual_participants, **age_months}, tsv_file)

    with pytest.warns(UserWarning, match='Multiple matches for participant_id:sub-2'):
        _get_age_from_tsv(tsv_file, 'participant_id', 'sub-2')



from pathlib import Path

from nibabies.interfaces.download import RetrievePoochFiles


def test_RetrievePoochFiles(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    getter = RetrievePoochFiles(intermediate='MNIInfant:cohort-1', target='MNI152NLin6Asym')
    outputs = getter.run().outputs
    assert Path(outputs.int2tgt_xfm).exists()
    assert Path(outputs.tgt2int_xfm).exists()

    cache = tmp_path / 'mycache'
    monkeypatch.setenv('NIBABIES_POOCH_DIR', cache)
    getter = RetrievePoochFiles(intermediate='MNIInfant:cohort-1', target='MNI152NLin6Asym')
    outputs = getter.run().outputs

    assert Path(outputs.int2tgt_xfm) == cache / 'from-MNIInfant+1_to-MNI152NLin6Asym_xfm.h5'
    assert Path(outputs.tgt2int_xfm) == cache / 'from-MNI152NLin6Asym_to-MNIInfant+1_xfm.h5'
